'''
Created on Aug 21, 2021

@author: immanueltrummer
'''
import math
import matplotlib
import matplotlib.pyplot as plt

def perf_breakdown(
        df, x_dim, x_vals, x_label, metric, 
        y_bounds, y_label, y_mode, out_path):
    """ Plots a multi-plot for a subset of approaches.
    
    Args:
        df: data frame containing all benchmark results
        x_dim: use this column for the x-axis
        x_vals: consider those values for x-axis
        x_label: use this label for x-axis
        metric: y axis summarizes this metric
        y_bounds: tuple of lower and upper y bound
        y_label: label of y axis
        y_mode: mode for y axis (e.g., linear or log)
        out_path: write plot to this file
    
    Returns:
        generated multi-plot
    """
    df = df.loc[:,[x_dim, 'nrfacts', 'nrpreds', metric]]
    df = df.query(f'{x_dim} in {x_vals}')
        
    _, axes = plt.subplots(
        nrows=3, ncols=3, figsize=(3.35,3), # 3.5, 3.5 
        subplotpars=matplotlib.figure.SubplotParams(
            wspace=0, hspace=0.75))
    
    for x_pos in range(3):
        for y_pos in range(3):
            nr_facts = x_pos + 1
            nr_preds = y_pos + 1
            ax = axes[y_pos, x_pos]
            if y_bounds is not None:
                ax.set_ylim(y_bounds[0], y_bounds[1])
            ax.set_title(f'{nr_facts}F, {nr_preds}P')
            if y_pos == 2:
                ax.set_xlabel(x_label)
            if x_pos == 0:
                ax.set_ylabel(y_label)

            q = f'nrfacts == {nr_facts} and nrpreds == {nr_preds}'
            cur_df = df.query(q)
            plot_data = []
            for approach in x_vals:
                app_data = cur_df.query(f'{x_dim} == "{approach}"')
                plot_data.append(app_data[metric])
                
            ax.boxplot(x=plot_data, labels=x_vals, showmeans=True)
            ax.yaxis.grid()
            ax.set_yscale(y_mode)

    #plt.subplots_adjust(wspace=0.3, hspace=0.4)
    plt.tight_layout(pad=1.05)
    plt.savefig(out_path)


def perf_plot(df, approaches, metric, y_bounds, y_label, y_mode, out_path):
    """ Draws aggregate data per approach and scenario.
    
    Args:
        df: data frame containing data to draw
        approaches: draw data for these approaches
        metric: draw this metric on the y axis
        y_bounds: tuple with lower and upper y bounds
        y_label: axis label for y axis
        y_mode: type of y axis (linear vs. log)
        out_path: write output plot to this file
    """
    print(f'Result type: {y_label}')
    _, axes = plt.subplots(nrows=2, ncols=2, figsize=(3,2.5))
    scenario_names = ['Laptops (30 KB)', 
                      'Tools (10 MB)',
                      'Flights (900 MB)',
                      'Sales (5 GB)']
    for scenario_id, scenario in enumerate([0, 1, 2, 3]):
        cur_df = df.query(f'scenario == {scenario}')
        df_pivot = cur_df.pivot(
            index=['testcase', 'nrfacts', 'nrpreds'], 
            columns='approach', values=metric)
        plot_data = [df_pivot[a] for a in approaches]
        approach_means = [(a, df_pivot[a].mean()) for a in approaches]
        print(f'Scenario {scenario}: {approach_means}')
        
        x_pos = math.floor(scenario_id / 2)
        y_pos = math.floor(scenario_id % 2)
        ax = axes[x_pos][y_pos]
        sc_name = scenario_names[scenario]
        ax.set_title(sc_name, fontsize=9)
        ax.set_ylabel(y_label)
        if y_bounds is not None:
            ax.set_ylim(y_bounds[0], y_bounds[1])
        ax.yaxis.grid()
        ax.set_yscale(y_mode)
        ax.boxplot(x=plot_data, labels=approaches, showmeans=True)
    
    plt.tight_layout(pad=1.05)
    #plt.subplots_adjust(hspace=0.7)
    plt.savefig(out_path)

def preprocess(df):
    """ Pre-process data frame containing CP results. 
    
    Args:
        df: data frame with CP results
    
    Returns:
        results with normalized approach and quality values
    """
    methods_short = {
        'rand1':'r', 'rand':'R', 'gen':'G', 'rlempty':'B',
        'sampleempty':'SB', 'sampleproactive':'SP', 
        'rlcube':'C', 'rlproactive':'P', 'rlNCproactive':'PN',
        'sample':'SP', 'simple':'SC', 'cluster':'CBR', 
        'bestof':'CP', 'viz':'V'}
    df['approach'] = df['approach'].apply(lambda a:methods_short[a])
    df['fquality'] = df['bquality'].apply(lambda q:max(q,-1))
    return df


'''
Created on Jul 23, 2021

@author: immanueltrummer
'''
# imports
import torch
from transformers import GPT2Tokenizer
from trl.gpt2 import GPT2HeadWithValueModel, respond_to_batch
from trl.ppo import PPOTrainer

# get models
gpt2_model = GPT2HeadWithValueModel.from_pretrained('gpt2')
gpt2_model_ref = GPT2HeadWithValueModel.from_pretrained('gpt2')
gpt2_tokenizer = GPT2Tokenizer.from_pretrained('gpt2')

# initialize trainer
ppo_config = {'batch_size': 1, 'forward_batch_size': 1}
ppo_trainer = PPOTrainer(gpt2_model, gpt2_model_ref, **ppo_config)

# encode a query
query_txt = "This morning I went to the "
query_tensor = gpt2_tokenizer.encode(query_txt, return_tensors="pt")

# get model response
response_tensor  = respond_to_batch(gpt2_model, query_tensor)
response_txt = gpt2_tokenizer.decode(response_tensor[0,:])

# define a reward for response
# (this could be any reward such as human feedback or output from another model)
reward = torch.tensor([1.0]) 

# train model with ppo
train_stats = ppo_trainer.step(query_tensor, response_tensor, reward)


'''
Created on May 5, 2023

@author: immanueltrummer
'''
import psycopg2.extras
import stable_baselines3
import time
import nminer.algs.rl
import nminer.sql.pred
import nminer.text.sum


def mine(
        db_name, db_user, db_pwd, table,
        preamble, dim_cols, dims_txt, agg_cols, aggs_txt, 
        target, nl_pattern, hg_model='facebook/bart-large-mnli',
        nr_facts=1, nr_preds=2, degree=5, nr_iterations=200):
    """
    Mines for facts that relate to natural language pattern.
    
    Args:
        db_name: name of Postgres database
        db_user: name of Postgres database user
        db_pwd: password of Postgres database user
        table: mine data in this table
        preamble: all facts start with this text
        dim_cols: names of categorical table columns for equality predicates
        dims_txt: list of text templates for predicates on dimension columns
        agg_cols: names of numerical table columns for aggregates
        aggs_txt: list of text templates for reporting aggregates
        target: SQL predicate describing target data compared to full data
        nl_pattern: search for facts that semantically relate to this pattern
        hg_model: ID of Huggingface Transformer model
        nr_facts: mine for relevant combinations of up to that many facts
        nr_preds: mine for facts using up to that many predicates
        degree: each search space node has that many neighbors
        nr_iterations: evaluate that many fact combinations
    
    Returns:
        a pair: a text reporting facts and a relevance score (w.r.t. pattern)
    """
    with psycopg2.connect(
        database=db_name, user=db_user, password=db_pwd,
        cursor_factory=psycopg2.extras.RealDictCursor) as connection:
        connection.autocommit = True
        
        # Mine table for facts related to pattern
        start_s = time.time()
        all_preds = nminer.sql.pred.all_preds(
            connection, table, dim_cols, target)
        sum_eval = nminer.text.sum.SumEvaluator(
            1, hg_model, nl_pattern)
        env = nminer.algs.rl.PickingEnv(
            connection, table, dim_cols, 
            agg_cols, [target], nr_facts, nr_preds,
            degree, nr_iterations, preamble, dims_txt, 
            aggs_txt, all_preds, c_type='proactive', 
            cluster=True, sum_eval=sum_eval)
        model = stable_baselines3.A2C(
            'MlpPolicy', env, verbose=True, 
            gamma=1.0, normalize_advantage=True)
        model.learn(total_timesteps=nr_iterations)
        total_s = time.time() - start_s
        print(f'Mining time: {total_s} seconds')
        
        # Find facts with highest relevance score
        rated_sums = env.s_eval.text_to_reward
        actual_sums = [i for i in rated_sums.items() if i[0] is not None]
        sorted_sums = sorted(actual_sums, key=lambda s: s[1])
        if sorted_sums:
            b_sum = sorted_sums[-1]
        else:
            b_sum = (None, -10)
    
        print(b_sum)
        return b_sum


'''
Created on May 25, 2022

@author: immanueltrummer
'''
import argparse
import matplotlib.pyplot as plt
import matplotlib.figure
import numpy as np
import pandas as pd
import statistics
from statsmodels.stats import inter_rater

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument('file', type=str, help='Path to AMT result file')
    args = parser.parse_args()
    
    df = pd.read_csv(args.file)
    group_cols = ['Input.Vizbest', 'Answer.semantic-similarity.label']
    all_counts_by_cat = df.groupby(by=group_cols).size().unstack()
    all_counts_by_cat.fillna(0, inplace=True)
    
    v_votes = []
    p_votes = []
    labels = []
    for scenario_short in ['L', 'D', 'F', 'S']:
        labels += [f'{scenario_short}{i}' for i in range(1,6)]
    
    all_agreements = []
    for scenario in [
        'Among all laptops', 'Among all developers', 
        'Among all flights', 'Among all liquors']:
        print(f'Scenario: {scenario}')
        counts_by_cat = all_counts_by_cat.filter(like=scenario, axis=0)
        
        print(f'Counts: {counts_by_cat}')
        nr_subjects = counts_by_cat.shape[0]
        agreements = []
        for i in range(nr_subjects):
            counts = counts_by_cat.iloc[i]
            v_votes.append(counts.iloc[0])
            p_votes.append(counts.iloc[1])
            agreement = inter_rater.fleiss_kappa([counts], method='unif')
            agreements.append(agreement)
            all_agreements.append(agreement)
        print(f'Agreement: {statistics.mean(agreements)}')
    
    print(f'All agreement: {statistics.mean(all_agreements)}')
    print(v_votes)
    print(p_votes)
    print(labels)
    
    width = 0.35  # the width of the bars
    
    plt.rcParams.update({'text.usetex': True, 'font.size':9,
                         'font.serif':['Computer Modern'],
                         'font.family':'serif'})
    plt.close('all')
    
    figure, axes = plt.subplots(
        nrows=1, ncols=4, figsize=(3.25,1.5))

    for i, title in enumerate(['Laptops', 'Developers', 'Flights', 'Sales']):
        ax = axes[i]
        ax.set_title(title)
        ax.get_xaxis().set_visible(False)
        ax.yaxis.grid()
        if i == 0:
            ax.set_ylabel('Votes')
        ax.set_ylim(0,15)
        x = np.arange(5)
        rects1 = ax.bar(x - width/2, v_votes[i*5:i*5+5], width, label='V')
        rects2 = ax.bar(x + width/2, p_votes[i*5:i*5+5], width, label='P', hatch='///')
        print('Output')
        print(f'Vizier: {sum(v_votes[i*5:i*5+5])}')
        print(f'BABOONS: {sum(p_votes[i*5:i*5+5])}')
        
    
    # Add some text for labels, title and custom x-axis tick labels, etc.
    # ax.set_ylabel('Votes')
    
    # ax.set_title('Scores by group and gender')
    # ax.set_xticks(x, labels)
    # ax.legend()
    
    # ax.yaxis.grid()
    # ax.bar_label(rects1, padding=3)
    # ax.bar_label(rects2, padding=3)
    
    figure.tight_layout(pad=1.01)
    
    # plt.show()
    plt.savefig('amt.pdf')


'''
Created on Aug 13, 2021

Analyzes results of summarization for single items.

@author: immanueltrummer
'''
import argparse
import nminer.stats.common
import matplotlib.pyplot as plt
import pandas as pd


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Analyze CP benchmark.')
    parser.add_argument('results_path', type=str, help='tsv file with results')
    args = parser.parse_args()
    
    df = pd.read_csv(args.results_path, sep='\t')
    df = nminer.stats.common.preprocess(df)
    print(df.info())

    plt.rcParams.update({'text.usetex': True, 'font.size':9,
                         'font.serif':['Computer Modern'],
                         'font.family':'serif'})
    plt.close('all')
    
    alg_groups = {'gen': ['r', 'R', 'G', 'V', 'P'], 'rl': ['B', 'P', 'SB', 'SP']}
    
    for scenario in range(4):
        scenario_data = df.query(f'scenario == {scenario}')
        path_prefix = f'plotsrev/S{scenario}_'
        
        for g_name, algs in alg_groups.items():
            file_path = path_prefix + f'{g_name}_qual.pdf'
            nminer.stats.common.perf_breakdown(
                scenario_data, 'approach', algs, 'Approach',
                'fquality', (-1.1, 1.1), 'Quality', 'linear', 
                file_path)
            
            file_path = path_prefix + f'{g_name}_time.pdf'
            nminer.stats.common.perf_breakdown(
                scenario_data, 'approach', algs, 'Approach', 
                'time', None, 'Time (s)', 'log', file_path)
    
    for g_name, algs in alg_groups.items():
        file_path = f'plotsrev/sx_{g_name}_time.pdf'
        nminer.stats.common.perf_plot(
            df, algs, 'time', None, 
            'Time (s)', 'log', file_path)
        file_path = f'plotsrev/sx_{g_name}_quality.pdf'
        nminer.stats.common.perf_plot(
            df, algs, 'fquality', (-1.1,1.1), 
            'Quality', 'linear', file_path)
    
    scenario_names = ['L', 'T', 'F', 'S']
    df['sc_name'] = df.apply(lambda r:scenario_names[r['scenario']], axis=1)
    
    df['relhits'] = df['chits'] / (df['chits'] + df['cmisses'])
    pro_hits = df.query('approach == "P"')
    hit_rate = pro_hits['relhits'].mean()
    print(f'Mean cache hit rate (proactive RL): {hit_rate}')
    nminer.stats.common.perf_breakdown(
        pro_hits, 'sc_name', scenario_names, 'Scenario',
        'relhits', (0, 1), 'Hit Ratio', 'linear', 'plotsrev/chits.pdf')
    
    df['relcost'] = df['nomergecost'] / (df['mergedcost'] + 0.001)
    sam_save = df.query('approach == "SB"')
    avg_save = sam_save['relcost'].mean()
    nminer.stats.common.perf_breakdown(
        sam_save, 'sc_name', scenario_names, 'scenario',
        'relcost', None, 'Savings', 'log', 'plotsrev/savings.pdf')
    print(f'Average relative savings: {avg_save}')


'''
Created on Jul 31, 2021

@author: immanueltrummer
'''
from nminer.cache.common import AggCache
from nminer.sql.query import AggQuery
import logging
import time

class DynamicCache(AggCache):
    """ Cache whose content is updated dynamically. """
    
    def __init__(self, connection, table, cmp_pred):
        """ Initializes proactive cache.
        
        Args:
            connection: connection to database
            table: name of source table
            cmp_pred: predicate used for comparisons
        """
        self.connection = connection
        self.table = table
        self.cmp_pred = cmp_pred
        self.q_to_r = {}
        self.q_to_c = {}
        self.c_hits = 0
        self.c_miss = 0

    def cache(self, g_query):
        """ Cache results for given group-by query. 
        
        Args:
            g_query: group-by query
        """
        sql = g_query.sql()
        logging.debug(f'About to fill cache with SQL "{sql}"')
        
        with self.connection.cursor() as cursor:
            start_s = time.time()
            cursor.execute(sql)
            total_s = time.time() - start_s
            logging.debug(f'Time: {total_s} s for query {sql}')
            rows = cursor.fetchall()
            self._extract_results(g_query, rows)

    def can_answer(self, query):
        """ Check if query result is cached.
        
        Args:
            query: look for this query's result
        
        Returns:
            true iff query result is cached
        """
        if query in self.q_to_r:
            self.c_hits += 1
            return True
        else:
            self.c_miss += 1
            return False
        
    def get_result(self, query):
        """ Get cached result for given query.
        
        Args:
            query: aggregation query for lookup
        
        Returns:
            result for given aggregation query
        """
        return self.q_to_r[query], self.q_to_c[query]

    def statistics(self):
        """ Return performance statistics.
        
        Returns:
            cache hits and misses in dictionary
        """
        return {'cache_hits':self.c_hits, 'cache_misses':self.c_miss}

    def update(self):
        """ Dynamic cache fills only upon specific request. """
        pass

    def _extract_results(self, g_query, rows):
        """ Extracts new cache entries from query result.
        
        Args:
            g_query: group-by query
            rows: result rows of caching query
        """
        for r in rows:
            for agg in g_query.aggs:
                cmp_c = r[f'cmp_c_{agg}']
                if cmp_c > 0:
                    c = r[f'c_{agg}']
                    s = r[f's_{agg}']
                    if s is not None and s > 0:
                        cmp_s = r[f'cmp_s_{agg}']
                        eq_preds = [(d, r[d]) for d in g_query.dims]
                        q = AggQuery(self.table, frozenset(eq_preds), 
                                     self.cmp_pred, agg)
                        rel_avg = (cmp_s/cmp_c)/(s/c)
                        self.q_to_r[q] = rel_avg
                        self.q_to_c[q] = cmp_c
                        
        for agg in g_query.aggs:
            for p_group in g_query.preds:
                q = AggQuery(self.table, frozenset(p_group), self.cmp_pred, agg)
                if not self.can_answer(q):
                    self.q_to_r[q] = None
                    self.q_to_c[q] = None


'''
Created on May 28, 2022

@author: immanueltrummer
'''
import argparse
import pandas as pd
import statistics
from statsmodels.stats import inter_rater

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument('path', type=str, help='Path to input file')
    args = parser.parse_args()
    
    df = pd.read_csv(args.path)
    print(f'DF columns: {df.columns}')
    
    
    nr_answers = df.groupby(by=['HITId']).size()
    print(nr_answers)
    print(f'Min answers : {min(nr_answers)}')
    print(f'Max answers: {max(nr_answers)}')
    
    groups = df.groupby(by=['HITId','Answer.semantic-similarity.label'])
    yn = groups.size().unstack()
    nr_topics = yn.shape[0]
    print(f'Nr topics: {nr_topics}')
    
    nr_yes = sum(yn.iloc[:,0] > yn.iloc[:,1])
    print(f'Nr. yes answers: {nr_yes}')
    
    agreements = []
    for i in range(nr_topics):
        y = yn.iloc[i,0]
        n = yn.iloc[i,1]
        agreement = inter_rater.fleiss_kappa([[y,n]], 'unif')
        agreements.append(agreement)
        
    mean_agreement = statistics.mean(agreements)
    print(f'Inter-rater agreement: {mean_agreement}')


'''
Created on Jul 21, 2021

@author: immanueltrummer
'''
from nminer.sql.pred import is_pred
import math
import random

class Fact():
    """ Represents one fact in a data summary. """
    
    def __init__(self, nr_preds):
        """ Initialize fact for given number of predicates. 
        
        Args:
            nr_preds: (maximal) number of scope predicates
        """
        self.nr_preds = nr_preds
        self.nr_props = nr_preds + 1
        self.reset()

    @staticmethod
    def from_props(props):
        """ Initializes new fact from properties.
        
        Args:
            props: fact properties
            
        Returns:
            new fact with given properties
        """
        nr_preds = len(props) - 1
        fact = Fact(nr_preds)
        for p_id, val in enumerate(props):
            fact.change(p_id, val)
        
        return fact

    def change(self, prop_id, new_val):
        """ Change fact property to new value. 

        Args:
            prop_id: change value of this property
            new_val: assign property to this value
        """
        self.props[prop_id] = new_val
        
    def get_agg(self):
        """ Returns index of aggregate. """
        return self.props[self.nr_preds]
    
    def get_id(self):
        """ Returns ID string. """
        return "_".join([str(p) for p in self.props])
    
    def get_preds(self):
        """ Returns fact predicates (indices). """
        return self.props[0:self.nr_preds]
    
    def get_prop(self, prop_id):
        """ Returns current value for property. """
        return self.props[prop_id]
    
    def is_agg(self, prop_id):
        """ Returns true iff property represents aggregate. """
        return True if prop_id == self.nr_preds else False
    
    def random_init(self, pred_cnt, agg_cnt):
        """ Initializes fact randomly.
        
        Args:
            pred_cnt: number of possible predicates
            agg_cnt: number of possible aggregates
        """
        for prop_id in range(self.nr_props):
            ub = agg_cnt if self.is_agg(prop_id) else pred_cnt
            self.props[prop_id] = random.randint(0, ub-1)
    
    def reset(self):
        """ Reset properties to default values. """
        self.props = [0] * self.nr_props
    
    def set_agg(self, value):
        """ Set aggregate to given value.
        
        Args:
            value: index of aggregation column
        """
        self.props[self.nr_preds] = value
    
    def set_pred(self, pred_idx, value):
        """ Set predicate to given value.
        
        Args:
            pred_idx: index of predicate to set
            value: set predicate to this value
        """
        self.props[pred_idx] = value
    
    def to_txt(self, preamble, dim_cols, all_preds, 
             dims_tmp, agg_cols, q_engine, aggs_txt):
        """ Generate text describing fact with confidence. 
        
        Args:
            preamble: start with this text
            dim_cols: list of dimension columns
            all_preds: list of all predicates
            dims_tmp: text templates for dimensions
            agg_cols: list of aggregation columns
            q_engine: used to calculate aggregates
            aggs_txt: text for aggregation columns
            
        Returns:
            text description of fact
        """
        f_parts = [preamble]
        preds = [all_preds[i] for i in self.get_preds()]
        preds = list(filter(lambda p:is_pred(p), preds))
        for pred in preds:
            dim_idx = dim_cols.index(pred[0])
            dim_tmp = dims_tmp[dim_idx]
            dim_txt = dim_tmp.replace('<V>', str(pred[1]))
            f_parts.append(dim_txt)
    
        agg_idx = self.get_agg()
        agg_col = agg_cols[agg_idx]
        rel_avg, row_cnt = q_engine.rel_avg(preds, agg_col)
        if rel_avg is None:
            return None, None
    
        percent = round(rel_avg * 100)
        percent_d = percent - 100
        if percent_d != 0:
            cmp_text = f'{abs(percent_d)}% '
            cmp_text += 'higher ' if percent_d > 0 else 'lower '
            cmp_text += 'than average'
        else:
            cmp_text = 'about average'                
        f_parts.append(f'{aggs_txt[agg_idx]} is {cmp_text}.')
        text = ' '.join(f_parts).replace('_', ' ').replace('  ', ' ')
        
        return text, self._confidence(row_cnt)
    
    def _confidence(self, row_cnt):
        """ Calculates confidence that text is accurate.
        
        For this implementation, we assume that all digits
        are used when generating text. Otherwise, we
        would need to consider the actual percentage.
        
        Args:
            row_cnt: rows considered for entity average

        Returns:
            confidence that text is accurate
        """
        return 2 * math.exp(-2*0.005**2/(row_cnt*1))


'''
Created on May 5, 2023

@author: immanueltrummer
'''
import nminer.interface.mine

# Name of Postgres database to mine
db_name = 'picker'
# User name for Postgres database
db_user = 'root'
# Password for Postgres database
db_pwd = ''
# Mine data in this table
table ='laptops'
# Table columns (typically categorical) for equality predicates
dim_cols = [
    'brand', 'processor_type', 'graphics_card', 'disk_space', 'ratings_5max']
# Text templates to express equality predicates associated with dimension
# columns. The <V> placeholder will be substituted by the predicate constant.
dims_txt = [
        'with <V> brand', 'with <V> CPU', 'with <V> graphics card', 
        'with <V> disk space', 'with <V> stars ']
# Table columns (typically numerical) for calculating aggregates.
agg_cols = ['display_size', 'discount_price', 'old_price']
# Text templates used to express aggregates on aforementioned columns.
aggs_txt = ['its display size', 'its discounted price', 'its old price']
# An SQL predicate describing the target data: we mine for facts that compare
# rows satisfying that predicate to all rows. Here, target data is associated
# with a specific laptop model and facts compare this to other laptops.
target = "laptop_name='VivoBook S430'"
# All facts start with this text, specifying what we compare to.
preamble = 'Among all laptops'
# Search for facts that comply with a pattern, described in natural language.
# E.g., we can search for arguments for or against buying this laptop.
nl_pattern = "arguments for buying this laptop"

# Mine for facts that are relevant, given the pattern
facts, relevance = nminer.interface.mine.mine(
    db_name, db_user, db_pwd, table,
    preamble, dim_cols, dims_txt, agg_cols, aggs_txt, 
    target, nl_pattern)

# Output mining results
print(f'Mined facts: {facts}')
print(f'Relevance: {relevance}')


'''
Created on Jun 5, 2021

@author: immanueltrummer
'''
from nminer.sql.pred import is_pred, pred_sql
from dataclasses import dataclass
from typing import FrozenSet, Tuple, Set

import logging
import time

@dataclass(frozen=True)
class AggQuery():
    """ Represents a simple aggregation query. """
    
    table: str
    eq_preds: FrozenSet[Tuple[str, str]]
    cmp_pred: str
    agg_col: str
    
    @staticmethod
    def from_fact(table, all_preds, cmp_pred, agg_cols, fact):
        """ Generate query associated with given fact.
        
        Args:
            table: fact refers to data in this table
            all_preds: list of all predicates
            cmp_pred: predicate for comparisons
            agg_cols: aggregation columns
            fact: construct query for this fact
        
        Returns:
            aggregation query
        """
        eq_preds = [all_preds[p] for p in fact.get_preds()]
        eq_preds = list(filter(lambda p:is_pred(p), eq_preds))
        eq_preds = frozenset(eq_preds)
        agg_idx = fact.get_agg()
        agg_col = agg_cols[agg_idx]
        return AggQuery(table, eq_preds, cmp_pred, agg_col)

    def pred_cols(self):
        """ Returns columns of query predicates. 
        
        Returns:
            frozen set with predicate columns
        """
        return frozenset([p[0] for p in self.eq_preds if is_pred(p)])


@dataclass
class GroupQuery():
    """ Represents a group-by aggregation query. """
    
    table: str
    dims: Set[str]
    cmp_pred: str
    aggs: Set[str]
    preds: Set[Tuple[str, str]]
    
    def __init__(self, table, dims, cmp_pred):
        """ Initializes group queries.
        
        Args:
            table: queries refer to this table
            dims: group by these columns
            cmp_pred: predicate for comparisons
        """
        self.table = table
        self.dims = dims
        self.cmp_pred = cmp_pred
        self.aggs = set()
        self.preds = set()
    
    @staticmethod
    def from_query(query):
        """ Initializes group-by query from simple aggregation query. 
        
        Args:
            query: simple aggregation query
        """
        g_q = GroupQuery(query.table, query.pred_cols(), query.cmp_pred)
        g_q.integrate(query)
        return g_q
    
    def can_merge(self, query):
        """ Check if simple query can be merged into group-by query.
        
        Args:
            query: simple aggregation query
        
        Returns:
            true iff the given query can be merged
        """
        return True if query.pred_cols == self.dims else False
    
    def contains(self, query):
        """ Check if group-by query generates result to simple query.
        
        Args:
            query: a simple aggregation query
        
        Returns:
            true iff the simple query is contained
        """
        if query.agg_col in self.aggs and query.eq_preds in self.preds and \
            query.table == self.table and query.cmp_pred == self.cmp_pred:
            return True
        else:
            return False
    
    def integrate(self, query):
        """ Expand group-by query to subsume given query.
        
        Args:
            query: integrate this aggregation query
        """
        assert(query.table == self.table)
        assert(query.pred_cols() == frozenset(self.dims))
        assert(query.cmp_pred == self.cmp_pred)
        
        self.aggs.add(query.agg_col)
        self.preds.add(query.eq_preds)
    
    def result_size(self):
        """ Calculates maximal result cardinality.
        
        Returns:
            Maximal result cardinality
        """
        return len(self.aggs) * len(self.preds)
    
    def sql(self):
        """ Generates equivalent SQL query.
        
        Returns:
            SQL representation of query
        """
        q_parts = [self._sql_select()]
        q_parts += [f' from {self.table}']
        q_parts += [self._sql_where()]
        q_parts += [self._sql_group()]
        return ' '.join(q_parts)

    def _sql_select(self):
        """ Generates select clause of cache query.
        
        Returns:
            SQL select clause for query
        """
        s_parts = []
        if self.dims:
            s_parts += list(self.dims)
        
        for agg_col in self.aggs:
            s_parts += [f'count({agg_col}) as c_{agg_col}']
            s_parts += [f'sum(case when {self.cmp_pred} ' \
                        f'and {agg_col} is not null then 1 ' \
                        f'else 0 end) as cmp_c_{agg_col}']
            s_parts += [f'sum({agg_col}) as s_{agg_col}']
            s_parts += [f'sum(case when {self.cmp_pred} then {agg_col} ' \
                        f'else 0 end) as cmp_s_{agg_col}']
        
        return 'select ' + ', '.join(s_parts)
    
    def _sql_group(self):
        """  Generates group-by clause of cache query.
            
        Returns:
            SQL group-by clause for query
        """
        if self.dims:
            return 'group by ' + ', '.join(self.dims)
        else:
            return ''
    
    def _sql_where(self):
        """ Generates where clause of cache query.
            
        Returns:
            SQL where clause for query
        """
        w_parts = []
        for p_group in self.preds:
            if p_group:
                c_pred = ' and '.join([pred_sql(p, v) for p, v in p_group])
                w_parts += ['(' + c_pred + ')']

        if w_parts:
            return ' where ' + ' or '.join(w_parts)
        else:
            return ''
 
    
class QueryEngine():
    """ Processes queries distinguishing entities from others. """
    
    def __init__(self, connection, table, cmp_pred, cache):
        """ Initialize query engine for specific connection.
        
        Args:
            connection: connection to database
            table: queries refer to this table
            cmp_pred: use for comparisons
            cache: cache for query results
        """
        self.connection = connection
        self.table = table
        self.cmp_pred = cmp_pred
        self.q_cache = cache
        self.nr_queries = 0
        
    def avg(self, eq_preds, pred, agg_col):
        """ Calculate average over aggregation column in scope. 
        
        Args:
            eq_preds: equality predicates as column-value pairs
            pred: SQL string representing predicate
            agg_col: calculate average for this column
            
        Returns:
            Average over aggregation column for satisfying rows, row count
        """
        q_parts = [f'select avg({agg_col}) as avg, ' \
                   f'count(*) as count from {self.table} where TRUE']
        q_parts += [pred_sql(col=c, val=v) for c, v in eq_preds]
        q_parts += [pred]
        query = ' AND '.join(q_parts)
        
        with self.connection.cursor() as cursor:
            logging.debug(f'About to execute query {query}')
            cursor.execute(query)
            row = cursor.fetchone()
            avg = row['avg']
            count = row['count']

        return avg, count
    
    def rel_avg(self, eq_preds, agg_col):
        """ Relative average of focus entity in given data scope. 
        
        Args:
            eq_preds: equality predicates defining scope
            agg_col: consider averages in this column
            
        Returns:
            Ratio of entity to general average
        """
        query = AggQuery(self.table, frozenset(eq_preds), 
                         self.cmp_pred, agg_col)
        
        if self.q_cache.can_answer(query):
            logging.debug(f'Cache hit: {query}')
            return self.q_cache.get_result(query)        
        else:
            logging.debug(f'Cache miss: {query}')
            self.nr_queries += 1
            start_s = time.time()
            e_avg, e_cnt = self.avg(eq_preds, self.cmp_pred, agg_col)
            g_avg, _ = self.avg(eq_preds, 'true', agg_col)
            total_s = time.time() - start_s
            logging.debug(f'Processing {query} took {total_s} seconds')
            
            if e_avg is None or g_avg is None:
                return None, None
            else:
                f_gen_avg = max(0.0001, float(g_avg))
                rel_avg = float(e_avg) / f_gen_avg
                return rel_avg, e_cnt
            
    def statistics(self):
        """ Generates performance statistics. 
        
        Returns:
            Dictionary with statistics
        """
        return {'nr_queries':self.nr_queries}


'''
Created on Jun 6, 2021

@author: immanueltrummer
'''
import argparse
import nminer.algs.base
import nminer.algs.sample
import nminer.eval.bench
import nminer.algs.rl
import nminer.sql.pred
import logging
import psycopg2
from psycopg2.extras import RealDictCursor
import time

from stable_baselines3 import A2C, PPO
from nminer.sql.pred import all_preds
from _collections import defaultdict

def print_details(env):
    """ Prints details about RL results.
    
    Args:
        env: extract info from this environment
    """
    best_summary = env.best_summary()
    print(f'Best summary: "{best_summary}"')
    for best in [True, False]:
        print(f'Top-K summaries (Best: {best})')
        topk_summaries = env.topk_summaries(2, best)
        for s in topk_summaries:
            print(s)
            
    print('Facts generated:')
    for f in env.fact_to_text.values():
        print(f'{f}')


def run_rl(connection, test_case, all_preds, nr_samples, c_type, cluster):
    """ Benchmarks primary method on test case.
    
    Args:
        connection: connection to database
        test_case: describes test case
        all_preds: ordered predicates
        c_type: type of cache to create
        cluster: whether to cluster search space
        
    Returns:
        summaries with reward, performance statistics
    """
    start_s = time.time()
    cmp_pred = test_case['cmp_pred']
    test_copy = test_case.copy()
    del test_copy['cmp_pred']
    test_copy['cmp_preds'] = [cmp_pred]
    env = nminer.algs.rl.PickingEnv(
        connection, **test_copy, all_preds=all_preds,
        c_type=c_type, cluster=cluster)
    model = A2C(
        'MlpPolicy', env, verbose=True, 
        gamma=1.0, normalize_advantage=True)
    model.learn(total_timesteps=nr_samples)
    total_s = time.time() - start_s
    logging.debug(f'Optimization took {total_s} seconds')
    
    p_stats = {'time':total_s}
    p_stats.update(env.statistics())
    
    return env.s_eval.text_to_reward, p_stats


def run_sampling(connection, test_case, all_preds, c_type):
    """ Run sampling algorithm. 

    Args:
        connection: connection to database
        test_case: summarize for this test case
        all_preds: all available predicates
        c_type: cache type to use for sampling
        
    Returns:
        summaries with quality, performance statistics
    """
    sampler = nminer.algs.sample.Sampler(
        connection, test_case, all_preds, 0.01, 5, c_type)
    text_to_reward, p_stats = sampler.run_sampling()
    
    return text_to_reward, p_stats


def run_random(connection, test_case, all_preds, nr_sums, timeout_s):
    """ Run simple random generation baseline.
    
    Args:
        connection: connection to database
        test_case: summarize for this test case
        all_preds: all available predicates
        nr_sums: generate so many summaries
        timeout_s: sample until this timeout
        
    Returns:
        summaries with quality, performance statistics
    """
    return nminer.algs.base.rand_sums(
        nr_sums=nr_sums, timeout_s=timeout_s, 
        connection=connection, all_preds=all_preds,
        **test_case)


def run_gen(connection, test_case, all_preds, timeout_s):
    """ Run generative learning baseline.
    
    Args:
        connection: connection to database
        test_case: summarize for this test case
        all_preds: all available predicates
        timeout_s: sample until this timeout
        
    Returns:
        summaries with quality, performance statistics
    """
    return nminer.algs.base.gen_rl(
        timeout_s=timeout_s, connection=connection, 
        all_preds=all_preds, **test_case)


def run_viz(connection, test_case, all_preds, timeout_s):
    """ Run baseline using Google Vizier platform. 
    
    Args:
        connection: connection to database
        test_case: describes summarization task
        all_preds: all available predicates
        timeout_s: time limit for iterations
    
    Returns:
        summaries with quality, performance statistics
    """
    return nminer.algs.base.vizier_sums(
        timeout_s=timeout_s, all_preds=all_preds,
        connection=connection, **test_case)


def log_line(outfile, b_id, t_id, nr_facts, 
             nr_preds, m_id, sums, p_stats):
    """ Writes one line to log file.
    
    Args:
        outfile: pointer to output file
        b_id: batch ID
        t_id: test case ID
        nr_facts: summaries use that many facts
        nr_preds: number of predicates per fact
        m_id: method identifier
        sums: maps summaries to rewards
        p_stats: performance statistics
    """
    actual_sums = [i for i in sums.items() if i[0] is not None]
    sorted_sums = sorted(actual_sums, key=lambda s: s[1])
    if sorted_sums:
        b_sum = sorted_sums[-1]
        w_sum = sorted_sums[0]
    else:
        b_sum = ('-', -10)
        w_sum = ('-', -10)
    
    def_stats = defaultdict(lambda:-1)
    def_stats.update(p_stats)
    e_time = def_stats['evaluation_time']
    time = def_stats['time']
    cache_hits = def_stats['cache_hits']
    cache_misses = def_stats['cache_misses']
    nr_queries = def_stats['nr_queries']
    no_merge_cost = def_stats['no_merge_cost']
    merged_cost = def_stats['merged_cost']

    outfile.write(f'{b_id}\t{t_id}\t{nr_facts}\t{nr_preds}\t' \
                  f'{m_id}\t{b_sum[0]}\t{b_sum[1]}\t{w_sum[0]}\t{w_sum[1]}\t' \
                  f'{time}\t{e_time}\t{cache_hits}\t{cache_misses}\t' \
                  f'{nr_queries}\t{no_merge_cost}\t{merged_cost}\n')
    outfile.flush()


def main():
    
    parser = argparse.ArgumentParser(description='Run CP benchmark.')
    parser.add_argument('db', type=str, help='database name')
    parser.add_argument('user', type=str, help='database user')
    parser.add_argument('out', type=str, help='result file name')
    parser.add_argument('log', type=str, help='logging level')
    parser.add_argument('samples', type=int, help='Number RL samples')
    args = parser.parse_args()

    db = args.db
    user = args.user
    outpath = args.out
    nr_samples = args.samples
    
    log_level = getattr(logging, args.log.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError(f'Invalid log level: {args.log}')
    logging.basicConfig(level=log_level, filemode='w')
    
    with open(outpath, 'w') as file:
        file.write('scenario\ttestcase\tnrfacts\tnrpreds\t'\
                   'approach\tbest\tbquality\tworst\twquality\t'\
                   'time\tetime\tchits\tcmisses\tnrqueries\t'\
                   'nomergecost\tmergedcost\n')
        with psycopg2.connect(database=db, user=user, 
                              cursor_factory=RealDictCursor) as connection:
            connection.autocommit = True
            
            test_batches = nminer.eval.bench.generate_testcases()
            for b_id, b in enumerate(test_batches):
                for t_id, t in enumerate(b):
                    all_preds = nminer.sql.pred.all_preds(
                        connection, t['table'], 
                        t['dim_cols'], t['cmp_pred'])
                    pred_cnt = len(all_preds)
                    print(f'Predicate search space size: {pred_cnt}')

                    for nr_facts in [1, 2, 3]:
                        for nr_preds in [1, 2, 3]:
                    
                            print(f'Next: B{b_id}/T{t_id}; ' \
                                  f'{nr_facts}F, {nr_preds}P')                            
                            t['nr_facts'] = nr_facts
                            t['nr_preds'] = nr_preds
                                                        
                            # sums, p_stats = run_rl(
                                # connection, t, all_preds, 
                                # nr_samples, 'proactive', False)
                            # log_line(
                                # file, b_id, t_id, nr_facts, nr_preds, 
                                # 'rlNCproactive', sums, p_stats)
                                
                            # for c_type in ['empty', 'proactive']:
                            for c_type in ['proactive']:
                                sums, p_stats = run_sampling(
                                    connection, t, all_preds, c_type)
                                log_line(
                                    file, b_id, t_id, nr_facts, nr_preds, 
                                    'sample' + c_type, sums, p_stats)
                                
                                sums, p_stats = run_rl(
                                    connection, t, all_preds, 
                                    nr_samples, c_type, True)
                                log_line(
                                    file, b_id, t_id, nr_facts, nr_preds, 
                                    'rl' + c_type, sums, p_stats)
                            timeout_s = p_stats['time']
                            
                            # sums, p_stats = run_random(
                                # connection, t, all_preds, 1, float('inf'))
                            # log_line(
                                # file, b_id, t_id, nr_facts,
                                # nr_preds, 'rand1', sums, p_stats)
                                #
                            # sums, p_stats = run_random(
                                # connection, t, all_preds, 
                                # float('inf'), timeout_s)
                            # log_line(
                                # file, b_id, t_id, nr_facts,
                                # nr_preds, 'rand', sums, p_stats)
                                #
                            # sums, p_stats = run_gen(
                                # connection, t, all_preds, timeout_s)
                            # log_line(
                                # file, b_id, t_id, nr_facts,
                                # nr_preds, 'gen', sums, p_stats)
                            
                            sums, p_stats = run_viz(
                                connection, t, all_preds, timeout_s)
                            log_line(
                                file, b_id, t_id, nr_facts,
                                nr_preds, 'viz', sums, p_stats)

if __name__ == '__main__':
    main()


'''
Created on Aug 15, 2021

@author: immanueltrummer
'''
import argparse
import nminer.algs.batch
import nminer.algs.sample
import nminer.sql.pred
import json
import logging
import psycopg2
from psycopg2.extras import RealDictCursor
import statistics
import time
import nminer.algs.batch
from nminer.algs.batch import IterativeClusters


def log_ic_results(nr_facts, nr_preds, method, avg_s, ic, out_file):
    """ Logs results to result file.
    
    Args:
        nr_facts: number of facts per summary
        nr_preds: number of predicates per fact
        method: string ID of method used
        avg_s: average seconds per item
        ic: iterative cluster processor
        out_file: write output to this file
    """
    for c_id, (_, b_evals) in ic.clusters.items():
        for pred, b_eval in b_evals.items():
            summary = b_eval.summary
            quality = b_eval.quality
            out_file.write(
                f'{nr_facts},{nr_preds},{method},{c_id},' +\
                f'{avg_s},"{pred}","{summary}",{quality}\n')


def log_si_results(nr_facts, nr_preds, method, iteration, avg_s, si, out_file):
    """ Log results of sub-modular, iterative algorithm.
    
    Args:
        nr_facts: number of facts
        nr_preds: number of predicates
        method: string ID of method used
        avg_s: average seconds per item
        si: sub-modular iterative optimizer
        out_file: handle to output file
    """
    for cmp_pred, s_eval in si.best_sums.items():
        out_file.write(
            f'{nr_facts},{nr_preds},{method},{iteration},' +\
            f'{avg_s},"{cmp_pred}","{s_eval.summary}",{s_eval.quality}\n')


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('in_path', type=str, help='Path to JSON input')
    parser.add_argument('db', type=str, help='Name of the database')
    parser.add_argument('user', type=str, help='Name of database user')
    parser.add_argument('out_path', type=str, help='Path to output files')
    parser.add_argument('log_level', type=str, help='Specify log level')
    args = parser.parse_args()
    
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError(f'Invalid log level: {args.log_level}')
    logging.basicConfig(level=log_level, filemode='w')
    
    with open(args.in_path) as in_file:
        batch = json.load(in_file)
        with open(args.out_path, 'w') as out_file:
            with psycopg2.connect(
                database=args.db, user=args.user, 
                cursor_factory=RealDictCursor) as connection:
                connection.autocommit = True

                all_preds = nminer.sql.pred.all_preds(
                    connection, batch['general']['table'], 
                    batch['general']['dim_cols'], 'true')                
                out_file.write(
                    'nrfacts,nrpreds,approach,iteration,' +\
                    'itemtime,pred,text,reward\n')

                for nr_facts, nr_preds in [
                        (3, 1), (3, 2), (3, 3),
                        (2, 1), (2, 2), (2, 3),
                        (1, 1), (1, 2), (1, 3),]:
                    batch['general']['nr_facts'] = nr_facts
                    batch['general']['nr_preds'] = nr_preds
    
                    nr_items = len(batch['predicates'])
                    # ic = nminer.algs.batch.IterativeClusters(
                        # connection, batch, all_preds)
                    si = nminer.algs.batch.SubModularIterative(
                        connection, batch, all_preds)
                    
                    si_start_s = time.time()
                    for i in range(5):
                        iteration_start_s = time.time()
                        logging.info(f'Starting batch iteration {i}')
                        # ic.iterate()
                        si.iterate()
                        total_s = time.time() - iteration_start_s
                        avg_s = total_s / nr_items
                        # log_ic_results(
                            # nr_facts, nr_preds, 'ic', 
                            # avg_s, ic, out_file)
                        log_si_results(
                            nr_facts, nr_preds, 'si', 
                            i, avg_s, si, out_file)
                    si_total_s = time.time() - si_start_s
                    
                    sample_start_s = time.time()
                    for cmp_pred in batch['predicates']:
                        test_case = batch['general'].copy()
                        test_case['cmp_pred'] = cmp_pred
                        
                        item_start_s = time.time()
                        sampler = nminer.algs.sample.Sampler(
                            connection, test_case, all_preds, 
                            0.01, 5, 'proactive')
                        text_to_reward, _ = sampler.run_sampling()
                        item_total_s = time.time() - item_start_s
                        
                        reward = max(text_to_reward.values())
                        reward = max(reward, -1)
                        text = max(text_to_reward.keys(), 
                            key=lambda k:text_to_reward[k])
                        out_file.write(
                            f'{nr_facts},{nr_preds},sample,0,{item_total_s},' +\
                            f'"{cmp_pred}","{text}",{reward}\n')
                            
                        if time.time() - sample_start_s > si_total_s:
                            break


'''
Created on Jun 16, 2021

@author: immanueltrummer
'''
def all_preds(connection, table, dim_cols, cmp_pred):
    """ Extracts all possible predicates from database. 
    
    Args:
        connection: connection to database
        table: extract predicates for this table
        dim_cols: list of dimension columns
        cmp_pred: filter rows by this predicate
    
    Returns:
        list of (column, value) pairs representing predicates    
    """
    preds = [("'any'", 'any')]
    for dim in dim_cols:
        print(f'Generating predicates for dimension {dim} ...')
        with connection.cursor() as cursor:
            query = f'select distinct {dim} from {table} ' \
                f'where {cmp_pred} and {dim} is not null'
            cursor.execute(query)
            result = cursor.fetchall()
            preds += [(dim, r[dim]) for r in result]
    return preds

def sql_esc(raw):
    """ Escapes SQL special symbols. 
    
    Args:
        raw: non-escaped SQL string
        
    Returns:
        escaped SQL string
    """
    return str(raw).replace("'", r"''")

def is_pred(pred):
    """ Checks if tuple represents predicate. 
    
    Args:
        pred: column and associated value
        
    Returns:
        true iff predicate represents condition
    """
    return not pred[0].startswith("'")

def pred_sql(col, val):
    """ Generates SQL for equality predicate.
    
    Args:
        col: predicate restricts this column
        val: filter column using this value
        
    Returns:
        predicate as SQL string (escaped)
    """
    esc_val = sql_esc(val)
    return f"{col}='{esc_val}'"


'''
Created on Aug 21, 2021

Analyzes results for approaches summarizing item batches.

@author: immanueltrummer
'''
import argparse
import nminer.stats.common
import matplotlib.pyplot as plt
import pandas as pd

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument('in_file', type=str, help='Path to input file')
    parser.add_argument('out_pre', type=str, help='Path prefix for output')
    args = parser.parse_args()
    
    df = pd.read_csv(args.in_file, sep=',')
    df['bquality'] = df['reward']
    df = nminer.stats.common.preprocess(df)
    df = df.loc[:,['nrfacts', 'nrpreds', 'approach', 'itemtime','fquality']]
    print(df.info())
    
    sum_agg = df.groupby(['nrfacts', 'nrpreds', 'approach'],
                         as_index=False).sum()
    avg_agg = df.groupby(['nrfacts', 'nrpreds', 'approach'],
                         as_index=False).mean()
    print(sum_agg)
    print(avg_agg)
    
    plt.rcParams.update({'text.usetex': True, 'font.size':9,
                         'font.serif':['Computer Modern'],
                         'font.family':'serif'})
    plt.close('all')
    
    nminer.stats.common.perf_breakdown(
        df, 'approach', ['SP', 'SC', 'CP'], 'Approach', 'itemtime', 
        None, 'Time (s)', 'log', f'{args.out_pre}timebreak.pdf')
    nminer.stats.common.perf_breakdown(
        df, 'approach', ['SP', 'SC', 'CP'], 'Approach', 'fquality',
        None, 'Quality', 'linear', f'{args.out_pre}qualbreak.pdf')


'''
Created on Jun 5, 2021

@author: immanueltrummer
'''
from collections import defaultdict
from nminer.sql.query import AggQuery
import nminer.cache.dynamic
import nminer.cache.static
import nminer.cache.proactive
from nminer.text.fact import Fact
from nminer.sql.pred import is_pred
from nminer.sql.query import QueryEngine
from nminer.text.sum import SumGenerator, SumEvaluator
from gym import spaces
from sentence_transformers import SentenceTransformer, util
import gym
import logging
import numpy as np
import statistics
import torch

class EmbeddingGraph():
    """ Graph connecting nodes with similar label embeddings. """

    def __init__(self, labels, degree, cluster):
        """ Generate graph with given labels.
        
        Args:
            labels: text labels for nodes
            degree: number of neighbors per node
            cluster: whether to cluster nodes by embedding
        """
        model = SentenceTransformer('paraphrase-MiniLM-L12-v2')
        self.embeddings = model.encode(labels, convert_to_tensor=True)
        cosine_sims = util.pytorch_cos_sim(self.embeddings, self.embeddings)
        self.neighbors = []
        
        nr_labels = len(labels)
        for i in range(nr_labels):
            prefs = cosine_sims[i,i:nr_labels]
            k = min(prefs.shape[0], degree)
            if cluster:
                _, indices = torch.topk(prefs, k)
                l_indices = indices.tolist()
            else:
                l_indices = list(range(i, i+k))
            l_indices += [0] * (degree - k)
            self.neighbors.append(l_indices)

    def get_embedding(self, node_id):
        """ Returns embedding of given node.
        
        Args:
            node_id: retrieve embedding for this node
            
        Returns:
            transformer embedding for given node
        """
        return self.embeddings[node_id]

    def get_neighbor(self, node, n_id):
        """ Retrieve specific neighbor of node. 
        
        Args:
            node: ID of node whose neighbors to examine
            n_id: index of node neighbor to retrieve
            
        Returns:
            i-th neighbor of j-th node
        """
        return self.neighbors[node][n_id]
    
    def get_reachable(self, start, steps):
        """ Retrieve all nodes reachable within a given number of steps. 
        
        Args:
            start: ID of start node
            steps: maximal number of steps
            
        Returns:
            all nodes reachable within given number of steps
        """
        reachable = set([start])
        for _ in range(steps):
            boundary = [nb for n in reachable for nb in self.neighbors[n]]
            reachable.update(boundary)
            
        return reachable

class PickingEnv(gym.Env):
    """ Environment for selecting facts for data summaries. """
    
    def __init__(self, connection, table, dim_cols, 
                 agg_cols, cmp_preds, nr_facts, nr_preds,
                 degree, max_steps, preamble, dims_tmp, 
                 aggs_txt, all_preds, c_type, cluster,
                 sum_eval=SumEvaluator(),
                 prior_best={}):
        """ Read database to initialize environment. 
        
        Args:
            connection: connection to database
            table: name of table to extract facts
            dim_cols: names of all property columns
            agg_cols: name of columns for aggregation
            cmp_preds: list of comparison predicates
            nr_facts: at most that many facts in description
            nr_preds: at most that many predicates per fact
            degree: degree for all transition graphs
            max_steps: number of steps per episode
            preamble: starts each data summary
            dims_tmp: assigns each dimension to text template
            aggs_txt: assigns each aggregate to text snippet
            all_preds: all possible predicates
            c_type: type of cache to create
            cluster: whether to cluster search space by embedding
            sum_eval: assigns summary text to quality values
            prior_best: maps predicates to quality of best prior summary
        """
        super(PickingEnv, self).__init__()
        self.connection = connection
        self.table = table
        self.dim_cols = dim_cols
        self.agg_cols = agg_cols
        self.cmp_preds = cmp_preds
        self.nr_items = len(cmp_preds)
        self.nr_facts = nr_facts
        self.nr_preds = nr_preds
        self.degree = degree
        self.max_steps = max_steps
        self.preamble = preamble
        self.dims_tmp = dims_tmp
        self.aggs_txt = aggs_txt
        
        self.agg_graph = EmbeddingGraph(agg_cols, degree, cluster)
        self.all_preds = all_preds
        pred_labels = [f'{p} is {v}' for p, v in self.all_preds]
        self.pred_graph = EmbeddingGraph(pred_labels, degree, cluster)
        
        self.proactive = True if c_type == 'proactive' else False
        self.caches = []
        self.q_engines = []
        self.s_gens = []
        for cmp_pred in cmp_preds:
            cache = self._create_cache(
                connection, table, dim_cols, 
                cmp_pred, agg_cols, nr_facts, 
                nr_preds, all_preds, c_type)
            q_engine = QueryEngine(
                connection, table, 
                cmp_pred, cache)
            s_gen = SumGenerator(
                all_preds, preamble, dim_cols, 
                dims_tmp, agg_cols, aggs_txt, 
                q_engine)
            self.caches.append(cache)
            self.q_engines.append(q_engine)
            self.s_gens.append(s_gen)

        self.s_eval = sum_eval
        self.prior_best = prior_best
        self.props_to_rewards = {}
        self.props_to_conf = {}
        
        self.cur_facts = []
        for _ in range(nr_facts):
            self.cur_facts.append(Fact(nr_preds))

        self.props_per_fact = nr_preds + 1
        action_dims = [nr_facts, self.props_per_fact, degree]
        self.action_space = spaces.MultiDiscrete(action_dims)
        
        self.nr_props = nr_facts * self.props_per_fact
        self.observation_space = spaces.Box(
            low=-10, high=10, shape=(self.nr_props, 384), dtype=np.float32)
    
        self.eval_s = 0
        self.nr_t_steps = 0
        self.reset()
        self._evaluate()

    def reset(self):
        """ Reset data summary to default. """
        self.nr_ep_steps = 0
        for fact in self.cur_facts:
            fact.reset()
            
        return self._observe()
    
    def statistics(self):
        """ Returns performance statistics. 
        
        Returns:
            Dictionary with performance statistics
        """
        stats = defaultdict(lambda:0)
        stats.update(self.s_eval.statistics())
        for cache in self.caches:
            stats.update(cache.statistics())
        for engine in self.q_engines:
            stats.update(engine.statistics())
        for gen in self.s_gens:
            stats.update(gen.statistics())
        
        return stats

    def step(self, action):
        """ Change fact or trigger evaluation. """
        self.nr_ep_steps += 1
        self.nr_t_steps += 1
        logging.info(
            f'Step {self.nr_t_steps} ' +\
            f'(step in episode: {self.nr_ep_steps})')
        
        if self.nr_ep_steps >= self.max_steps:
            done = True
            reward = 0
        else:
            fact_idx = action[0]
            prop_idx = action[1]
            nb_idx = action[2]
            fact = self.cur_facts[fact_idx]
            cur_val = fact.get_prop(prop_idx)
            if fact.is_agg(prop_idx):
                new_val = self.agg_graph.get_neighbor(cur_val, nb_idx)
            else:
                new_val = self.pred_graph.get_neighbor(cur_val, nb_idx)
            self.cur_facts[fact_idx].change(prop_idx, new_val)
            
            done = False
            for cache in self.caches:
                if self.proactive:
                    cache.set_cur_facts(self.cur_facts)
                cache.update()
            reward = self._evaluate()
        
        return self._observe(), reward, done, {}

    def _create_cache(self, connection, table, dim_cols, cmp_pred, 
                      agg_cols, nr_facts, nr_preds, all_preds, c_type):
        """ Creates a cache object as specified.
        
        Args:
            connection: connection to database
            table: cache stores results of queries on this table
            dim_cols: dimension columns
            cmp_pred: compare data subsets defined by this predicate
            agg_cols: aggregation columns
            nr_facts: summary contains so many facts
            nr_preds: maximal number of predicates per fact
            all_preds: all possible predicates on dimensions
            c_type: string identifier of cache type
        Returns:
            newly created cache object
        """
        if c_type == 'dynamic':
            cache = nminer.cache.dynamic.DynamicCache(connection, table, cmp_pred)
        elif c_type == 'empty':
            cache = nminer.cache.static.EmptyCache()
        elif c_type == 'cube':
            cache = nminer.cache.static.CubeCache(
                connection, table, dim_cols, cmp_pred, agg_cols, 900)
        elif c_type == 'proactive':
            cache = nminer.cache.proactive.ProCache(
                connection, table, cmp_pred, nr_facts, 
                nr_preds, all_preds, self.pred_graph,
                agg_cols, self.agg_graph)
        else:
            raise ValueError(f'Unknown cache type: {c_type}')
        return cache

    def _evaluate(self):
        """ Evaluate quality of current summary. """
        confs = []
        rewards = []
        for cmp_pred, gen in zip(self.cmp_preds, self.s_gens):
            text, conf = gen.generate(self.cur_facts)
            this_quality = self.s_eval.evaluate(text)
            prior_quality = self.prior_best.get(cmp_pred)
            reward = this_quality
            if prior_quality is not None:
                reward -= prior_quality
                reward = max(reward, 0)
            rewards.append(reward)
            logging.debug(
                f'Reward: {reward} for {text}; ' +\
                f'prior quality: {prior_quality}')
            if conf is not None:
                confs.append(conf)
        
        mean_conf = statistics.mean(confs) if confs else 0
        mean_reward = statistics.mean(rewards)
        logging.debug(f'Mean reward {mean_reward}')
        self._save_eval_results(mean_reward, mean_conf)
        return mean_reward

    def _observe(self):
        """ Returns observations for learning agent. """
        components = []
        for fact_ctr in range(self.nr_facts):
            fact = self.cur_facts[fact_ctr]
            preds = fact.get_preds()
            for pred_idx in preds:
                pred_emb = self.pred_graph.get_embedding(pred_idx)
                components.append(pred_emb)
                
            agg_idx = fact.get_agg()
            agg_emb = self.agg_graph.get_embedding(agg_idx)
            components.append(agg_emb)
        
        return torch.stack(components, dim=0).to('cpu').numpy()
    
    def _save_eval_results(self, reward, conf):
        """ Store reward of current fact combination. 
        
        Args:
            reward: received for current facts
            conf: confidence of evaluation (used if sampling)
        """
        fact_tuples = [tuple(f.props) for f in self.cur_facts]
        sum_tuple = tuple(fact_tuples)
        self.props_to_rewards[sum_tuple] = reward
        self.props_to_conf[sum_tuple] = conf


'''
Created on Jul 31, 2021

@author: immanueltrummer
'''
from abc import ABC, abstractmethod

class AggCache(ABC):
    """ Common superclass for all cache types. """
    
    @abstractmethod
    def can_answer(self, query):
        """ Checks if cache contains answer to query.
        
        Args:
            query: search for result of this query
            
        Returns:
            true iff cache answers query
        """
        raise NotImplementedError

    @abstractmethod
    def get_result(self, query):
        """ Get result for query from cache content.
        
        Args:
            query: retrieve result of this query
            
        Returns:
            pair of relative average and, optionally, row count
        """
        raise NotImplementedError

    @abstractmethod
    def statistics(self):
        """ Get statistics on cache performance.
        
        Returns:
            dictionary with performance statistics
        """
        raise NotImplementedError

    @abstractmethod
    def update(self):
        """ Update cache content (no effect for static cache). """
        raise NotImplementedError


'''
Created on Jul 22, 2021

@author: immanueltrummer
'''
import logging
import time
import torch
import transformers


class SumGenerator():
    """ Class for generating summaries. """
    
    def __init__(self, all_preds, preamble, dim_cols, 
                 dims_tmp, agg_cols, aggs_txt, q_engine):
        """ Initializes parameters and fact cache. 
        
        Args:
            all_preds: all relevant predicates
            preamble: starts each fact text
            dim_cols: dimension columns
            dims_tmp: dimension templates
            agg_cols: aggregation columns
            aggs_txt: text describing column
            q_engine: used for querying
        """
        self.all_preds = all_preds
        self.preamble = preamble
        self.dim_cols = dim_cols
        self.dims_tmp = dims_tmp
        self.agg_cols = agg_cols
        self.aggs_txt = aggs_txt
        self.q_engine = q_engine
        self.fact_to_text = {}
        self.fact_to_conf = {}
        self.gen_s = 0
    
    def generate(self, facts):
        """ Generate text describing given facts. 
        
        Args:
            facts: describe those facts
            
        Returns:
            text string describing facts, confidence
        """
        start_s = time.time()
        s_parts = []
        s_conf = 1
        for fact in facts:
            f_id = fact.get_id()
            if f_id in self.fact_to_text:
                f_txt = self.fact_to_text[f_id]
                f_conf = self.fact_to_conf[f_id]
            else:
                f_txt, f_conf = fact.to_txt(
                    preamble=self.preamble, dim_cols=self.dim_cols, 
                    all_preds=self.all_preds, dims_tmp=self.dims_tmp, 
                    agg_cols=self.agg_cols, q_engine=self.q_engine, 
                    aggs_txt=self.aggs_txt)
                self.fact_to_text[f_id] = f_txt
                self.fact_to_conf[f_id] = f_conf
                
            if f_txt is None:
                return None, None
            s_parts.append(f_txt)
            s_conf *= f_conf
    
        self.gen_s += time.time() - start_s
        return ' '.join(s_parts), s_conf
    
    def statistics(self):
        """ Returns performance statistics.
        
        Returns:
            Dictionary with statistics
        """
        return {'generation_time':self.gen_s}


class SumEvaluator():
    """ Class for evaluating data summaries. """
    
    def __init__(self, goal=0, 
                 model_name='siebert/sentiment-roberta-large-english', 
                 label=''):
        """ Initializes cache, statistics, and model. 
        
        Args:
            goal: optimization goal (0 for default, 1 for customized)
            model_name: name of Huggingface model used for text evaluation
            label: compare text to this label for custom goals
        """
        self.goal = goal
        self.model_name = model_name
        self.label = label
        self.model = self._init_model()
        self.eval_s = 0
        self.text_to_reward = {}
    
    def evaluate(self, text):
        """ Evaluate quality of current summary. 
        
        Args:
            text: summary text to evaluate
            
        Returns:
            numerical value (higher is more positive)
        """
        start_s = time.time()
        if text is None:
            reward = -10
        else:
            if text in self.text_to_reward:
                reward = self.text_to_reward[text]
            else:
                reward = self._evaluate_text(text)
                self.text_to_reward[text] = reward
    
            logging.debug(f'Reward {reward} for "{text}"')
        
        self.eval_s += time.time() - start_s
        return reward
    
    def evaluate_batch(self, texts):
        """ Evaluate batch of summaries.
        
        Args:
            texts: list of text summaries to evaluate
        
        Returns:
            list of associated scores
        """
        texts = [t if t is not None else '' for t in texts]
        nr_texts = len(texts)
        n = 1000
        batches = [texts[i:i+n] for i in range(0, nr_texts, n)]
        
        results = []
        for batch in batches:
            results += self.model(batch, batch_size=8)
        scores = [self._extract_score(r) for r in results]
        for idx, text in enumerate(texts):
            if not text:
                scores[idx] = -1
        return scores
    
    def statistics(self):
        """ Returns performance statistics. 
        
        Returns:
            dictionary with performance statistics
        """
        return {'evaluation_time':self.eval_s}
    
    def _init_model(self):
        """ Initialize model used for text evaluation.
        
        Returns:
            newly initialized model for evaluation
        """
        if self.goal == 0:
            task = 'sentiment-analysis'
        elif self.goal == 1:
            task = 'zero-shot-classification'
        else:
            raise ValueError(f'Error - unknown goal code: {self.goal}')
        
        if torch.cuda.is_available():
            device = torch.cuda.current_device()
        else:
            device = -1
        return transformers.pipeline(
            task, model=self.model_name, 
            device=device)
    
    def _evaluate_text(self, text):
        """ Evaluate quality of input text. 
        
        Returns:
            quality value between -1 (or 0) and +1 (higher is better)
        """
        if self.goal == 0:
            result = self.model(text)[0]
        elif self.goal == 1:
            result = self.model(
                text, [self.label])
        
        return self._extract_score(result)
    
    def _extract_score(self, result):
        """ Extract scores from model results.
        
        Args:
            result: result of model evaluation
        
        Returns:
            score representing text quality
        """
        if self.goal == 0:
            label = result['label']
            score = result['score']
            if label == 'POSITIVE':
                return score
            else:
                return -score
        
        elif self.goal == 1:
            return result['scores'][0]


'''
Created on Aug 18, 2021

@author: immanueltrummer
'''
from nminer.cache.common import AggCache
from nminer.sql.pred import is_pred, pred_sql
from nminer.sql.query import AggQuery
from nminer.text.fact import Fact
import logging

class MultiItemCache(AggCache):
    """ Proactively caches results for multiple items. """
    
    def __init__(self, connection, table, cmp_col, 
                 all_preds, agg_cols, sum_tmp):
        """ Proactively caches results for summary template for all items. 
        
        Args:
            connection: connection to database
            table: execute queries on this table
            cmp_col: each item associated with one value here
            all_preds: all predicates on dimension columns
            agg_cols: aggregation columns
            sum_tmp: list of fact properties
        """
        self.q_to_r = {}
        with connection.cursor() as cursor:
            for f_idx, props in enumerate(sum_tmp):
                logging.info(f'Evaluating fact {f_idx} ...')
                q_parts = [f'select {cmp_col} as cmp_col,']
                fact = Fact.from_props(props)
                agg_idx = fact.get_agg()
                agg = agg_cols[agg_idx]
                q_parts.append(
                    f'avg({agg})/(select avg({agg}) ' + \
                    f'from {table}) as rel_avg')
                
                q_parts.append(f'from {table} where')
                w_parts = ['true']
                for p_idx in fact.get_preds():
                    pred = all_preds[p_idx]
                    if is_pred(pred):
                        col, val = pred
                        w_parts.append(pred_sql(col, val))

                w_clause = ' and '.join(w_parts)
                q_parts.append(w_clause)
                q_parts.append(f'group by {cmp_col}')
                sql = ' '.join(q_parts)

                logging.debug(f'Multi-cache SQL: {sql}')
                cursor.execute(sql)
                rows = cursor.fetchall()
                for row in rows:
                    cmp_val = row['cmp_col']
                    rel_avg = row['rel_avg']
                    cmp_pred = pred_sql(cmp_col, cmp_val)
                    agg_q = AggQuery.from_fact(
                        table, all_preds, cmp_pred, 
                        agg_cols, fact)
                    self.q_to_r[agg_q] = rel_avg
                #
                # # Determine items with empty data sets
                # sql = f'select distinct {cmp_col} as cmp_col from {table}'
                # cursor.execute(sql)
                # rows = cursor.fetchall()
                # for row in rows:
                    # cmp_val = row['cmp_col']
                    # cmp_pred = pred_sql(cmp_col, cmp_val)
                    # agg_q = AggQuery.from_fact(
                        # table, all_preds, cmp_pred, 
                        # agg_cols, fact)
                    # if agg_q not in self.q_to_r:
                        # self.q_to_r[agg_q] = None
            
        logging.debug(f'Multi-item cache content: {self.q_to_r}')
    
    def can_answer(self, _):
        """ Un-cached queries return None value by default. """
        return True
    
    def get_result(self, query):
        return self.q_to_r.get(query), -1
    
    def statistics(self):
        return {'cache_hits':-1, 'cache_misses':-1}
    
    def update(self):
        pass


'''
Created on Aug 7, 2021

@author: immanueltrummer
'''
from nminer.sql.query import AggQuery, GroupQuery
import nminer.cache.dynamic
import nminer.text.fact
import nminer.algs.rl
import nminer.sql.cost
import nminer.text.sum
import logging
from stable_baselines3 import A2C
import time

def create_sample(connection, table, sample_ratio):
    """ Creates table containing sample from source.
    
    Args:
        connection: connection to database
        table: name of sample source table
        sample_ratio: ratio of samples
    
    Returns:
        name of table containing sample
    """
    total_rows, _ = nminer.sql.cost.estimates(
        connection, f'select * from {table}')
    sample_rows = int(
        max(10000, total_rows * sample_ratio))
    
    sample_tbl = f'{table}_sample'
    with connection.cursor() as cursor:
        cursor.execute(f'drop table if exists {sample_tbl}')
        cursor.execute(f'create unlogged table {sample_tbl} as ' \
                       f'(select * from {table} limit {sample_rows})')
    
    return sample_tbl


class Sampler():
    """ Selects data summaries via sampling. """
    
    def __init__(self, connection, test_case, all_preds, 
                 sample_ratio, max_nr_sums, c_type):
        """ Initialize sampler, creates summary generator.
        
        Args:
            connection: connection to database
            test_case: description of scenario
            all_preds: all possible predicates
            sample_ratio: sample at least this ratio of the rows
            max_nr_sums: maximal number of summaries on full data set
            c_type: cache type used for sampling
        """
        self.connection = connection
        self.test_case = test_case
        self.all_preds = all_preds
        self.sample_ratio = sample_ratio
        self.max_nr_sums = max_nr_sums
        self.c_type = c_type
        self.table = test_case['table']
        self.cmp_pred = test_case['cmp_pred']
        self.dim_cols = test_case['dim_cols']
        self.agg_cols = test_case['agg_cols']
        preamble = test_case['preamble']
        dims_tmp = test_case['dims_tmp']
        aggs_txt = test_case['aggs_txt']
        
        self.cache = nminer.cache.dynamic.DynamicCache(
            connection, self.table, self.cmp_pred)
        self.q_engine = nminer.sql.query.QueryEngine(
            connection, self.table, 
            self.cmp_pred, self.cache)
        self.s_gen = nminer.text.sum.SumGenerator(
            all_preds, preamble, self.dim_cols,
            dims_tmp, self.agg_cols, aggs_txt,
            self.q_engine)
        self.s_eval = nminer.text.sum.SumEvaluator()

    def _cost(self, g_queries):
        """ Calculates cost estimate for a set of group-by queries.
        
        Args:
            g_queries: group-by queries
        
        Returns:
            accumulated cost for set of group-by queries
        """
        total_cost = 0
        for g_q in g_queries:
            sql = g_q.sql()
            _, cost = nminer.sql.cost.estimates(self.connection, sql)
            total_cost += cost
        
        return total_cost

    
    def _fill_cache(self, sam_sums):
        """ Fill cache with query results needed to generate summaries.
        
        Args:
            sam_sums: summaries (facts) selected via sampling
        
        Returns:
            statistics on efficiency of merge step
        """
        logging.debug(f'Filling cache for summaries: {sam_sums}')
        stats = {}
        # collect queries associated with facts
        queries = []
        for sum_facts in sam_sums:
            for fact in sum_facts:
                a_q = AggQuery.from_fact(
                    self.table, self.all_preds,
                    self.cmp_pred, self.agg_cols, fact)
                queries += [a_q]
        
        logging.debug(f'Raw queries to cache: {queries}')
        
        # merge queries with same aggregates and dimensions
        g_merged = []
        for a_query in queries:
            is_merged = False
            for merged in g_merged:
                if merged.can_merge(a_query) and \
                a_query.agg_col in merged.agg_cols:
                    merged.integrate(a_query)
                    is_merged = True
                    break
            
            if not is_merged:
                g_query = GroupQuery.from_query(a_query)
                g_merged += [g_query]

        # merge queries with same dimensions until cost increases
        min_g_qs = g_merged.copy()
        min_cost = self._cost(g_merged)
        logging.debug(f'Cost before clustering: {min_cost} with {g_merged}')
        stats['no_merge_cost'] = min_cost
        
        improved = True
        while improved:
            sims = []
            for g1 in g_merged:
                for g2 in g_merged:
                    if not (g1 == g2):
                        sim = self._similarity(g1, g2)
                        sims += [(g1, g2, sim)]
            
            improved = False
            if sims:
                g1, g2, _ = max(sims, key=lambda s:s[2])
                gm = GroupQuery(self.table, g1.dims, self.cmp_pred)
                gm.preds.update(g1.preds)
                gm.preds.update(g2.preds)
                gm.aggs.update(g1.aggs)
                gm.aggs.update(g2.aggs)
                g_merged.remove(g1)
                g_merged.remove(g2)
                g_merged.append(gm)
                
                cost = self._cost(g_merged)
                if cost < min_cost:
                    min_cost = cost
                    min_g_qs = g_merged.copy()
                    improved = True
        
        logging.debug(f'Cost after clustering: {min_cost} with {min_g_qs}')
        stats['merged_cost'] = min_cost
        for g_q in min_g_qs:
            self.cache.cache(g_q)
        
        return stats
    
    def _full_data_summaries(self, sam_sums):
        """ Generates summaries by considering the full data set.
        
        Args:
            sam_sums: summaries selected via sample
        
        Returns:
            a subset of summaries referring to full data set, statistics
        """
        text_to_reward = {}
    
        nr_sums = min(len(sam_sums), self.max_nr_sums)
        sel_sums = sam_sums[0:nr_sums]
        if sel_sums:
            stats = self._fill_cache(sel_sums)
        else:
            stats = {}
        
        for facts in sel_sums:
            queries = []
            for fact in facts:
                query = AggQuery.from_fact(
                    self.table, self.all_preds, 
                    self.cmp_pred, self.agg_cols, fact)
                queries.append(query)
            
            if not [q for q in queries if not self.cache.can_answer(q)]:
                text, _ = self.s_gen.generate(facts)
                reward = self.s_eval.evaluate(text)
                text_to_reward[text] = reward
        
        return text_to_reward, stats

    def _learn_on_sample(self):
        """ Learn good summaries for a data sample.
        
        Returns:
            summaries sorted by estimated quality (descending), statistics
        """
        sample_case = self.test_case.copy()
        table_sample = create_sample(
            self.connection, self.table, 
            self.sample_ratio)
        sample_case['table'] = table_sample
        cmp_pred = sample_case['cmp_pred']
        del sample_case['cmp_pred']
        sample_case['cmp_preds'] = [cmp_pred]
        env = nminer.algs.rl.PickingEnv(
            self.connection, **sample_case, 
            all_preds=self.all_preds, 
            c_type=self.c_type, cluster=True)
        model = A2C(
            'MlpPolicy', env, verbose=True, 
            gamma=1.0, normalize_advantage=True)
        model.learn(total_timesteps=200)
        
        return self._ranked_sums(env), env.statistics()
    
    def _ranked_sums(self, env):
        """ Extracts summaries from environment and ranks them.
        
        Args:
            env: reinforcement learning environment
        
        Returns:
            list of summaries in decreasing order of preference
        """
        prop_sums = env.props_to_rewards.keys()
        sum_eval = []
        for prop_sum in prop_sums:
            reward = env.props_to_rewards[prop_sum]
            conf = env.props_to_conf[prop_sum]
            if reward is not None and conf is not None:
                c_qual = reward * conf
                fact_sum = [nminer.text.fact.Fact.from_props(p) for p in prop_sum]
                sum_eval += [(fact_sum, c_qual)]
        
        s_sums = sorted(sum_eval, key=lambda s_e: s_e[1], reverse=True)
        return [s[0] for s in s_sums]
            
    def _similarity(self, g_query_1, g_query_2):
        """ Calculates similarity (from 0 to 1) between two queries.
        
        Args:
            g_query_1: first group-by query
            g_query_2: second group-by query
        
        Returns:
            similarity between 0 and 1
        """
        if g_query_1.dims == g_query_2:
            preds_1 = g_query_1.preds.copy()
            preds_2 = g_query_2.preds.copy()
            c_union = len(preds_1.union(preds_2))
            c_inter = len(preds_1.intersect(preds_2))
            return c_inter / c_union
        else:
            return 0
                
    def run_sampling(self):
        """ Run sampling algorithm. 
        
        Returns:
            summaries with quality, performance statistics
        """
        start_s = time.time()
        s_sums, p_stats = self._learn_on_sample()
        fs_to_r, m_stats = self._full_data_summaries(s_sums)
        total_s = time.time() - start_s
        
        logging.debug(f'Optimization took {total_s} seconds')
        p_stats['time'] = total_s
        p_stats.update(m_stats)
        
        return fs_to_r, p_stats


'''
Created on Jul 21, 2021

@author: immanueltrummer
'''
import nminer.algs.vizier
import nminer.cache.static
import nminer.text.fact
import nminer.sql.query
import nminer.text.sum
import time
import torch

from transformers import AutoTokenizer, AutoModelForSequenceClassification
from transformers import GPT2Tokenizer
from trl.gpt2 import GPT2HeadWithValueModel, respond_to_batch
from trl.ppo import PPOTrainer

# Prepare extractive summarization
mnli_tokenizer = AutoTokenizer.from_pretrained('roberta-large-mnli')
mnli_model = AutoModelForSequenceClassification.from_pretrained(
    'roberta-large-mnli')
mnli_model.eval()


def rand_sums(
        nr_sums, timeout_s, connection, table, dim_cols, 
        agg_cols, cmp_pred, nr_facts, nr_preds, preamble, 
        dims_tmp, aggs_txt, all_preds, **kwargs):
    """ Generates and evaluates one random summary.
    
    Args:
        nr_sums: max. number of summaries to generate
        timeout_s: max. time in seconds
        connection: connects to database
        table: summarize this table
        dim_cols: dimension columns
        agg_cols: aggregation columns
        cmp_pred: identifies entity to advertise
        nr_facts: max. number of facts in summary
        nr_preds: max. number of predicates per fact
        preamble: start summary with this text
        dims_tmp: text templates for dimensions
        aggs_txt: text describing aggregates
        all_preds: all possible predicates
    
    Returns:
        Dictionary mapping summaries to reward, statistics
    """
    start_s = time.time()
    cache = nminer.cache.static.EmptyCache()
    q_engine = nminer.sql.query.QueryEngine(
        connection, table, cmp_pred, cache)
    s_gen = nminer.text.sum.SumGenerator(
        all_preds, preamble, dim_cols, dims_tmp, 
        agg_cols, aggs_txt, q_engine)
    s_eval = nminer.text.sum.SumEvaluator()

    pred_cnt = len(all_preds)
    agg_cnt = len(agg_cols)
    text_to_quality = {}
    
    counter = 0
    while counter < nr_sums:
        counter += 1
        
        facts = []
        for _ in range(nr_facts):
            fact = nminer.text.fact.Fact(nr_preds)
            fact.random_init(pred_cnt=pred_cnt, agg_cnt=agg_cnt)
            facts.append(fact)
    
        text, _ = s_gen.generate(facts)
        quality = s_eval.evaluate(text)
        text_to_quality[text] = quality
        
        total_s = time.time() - start_s
        if total_s > timeout_s:
            break
    
    stats = {'time':total_s}
    stats.update(cache.statistics())
    stats.update(q_engine.statistics())
    stats.update(s_gen.statistics())
    stats.update(s_eval.statistics())
    
    return text_to_quality, stats


def escape_text(raw_text):
    """ Escape generated text.
    
    Args:
        raw_text: text potentially containing double quotes
        
    Returns:
        text with escaped double quotes
    """
    return raw_text.replace('"', '""')


def gen_rl(timeout_s, **kwargs):
    """ Generative baseline using reinforcement learning. 
    
    Args:
        timeout_s: timeout in seconds
        kwargs: dictionary with arguments
        
    Returns:
        Dictionary mapping summaries to quality values
    """
    gpt2_tokenizer = GPT2Tokenizer.from_pretrained('gpt2')
    gpt2_model = GPT2HeadWithValueModel.from_pretrained('gpt2')
    gpt2_model_ref = GPT2HeadWithValueModel.from_pretrained('gpt2')
    
    ppo_config = {'batch_size': 1, 'forward_batch_size': 1}
    ppo_trainer = PPOTrainer(gpt2_model, gpt2_model_ref, **ppo_config)
    
    r_params = kwargs.copy()
    r_params['nr_facts'] = 10
    r_params['nr_preds'] = 1
    s_eval = nminer.text.sum.SumEvaluator()
    start_s = time.time()
    t_to_q = {}
    eval_total_s = 0
    
    while True:
        rt_to_q, _ = rand_sums(1, float('inf'), **r_params)
        
        rand_facts = list(rt_to_q.keys())[0]
        if rand_facts is not None:
            prompt_pt = gpt2_tokenizer.encode(rand_facts, return_tensors="pt")        
            sum_pt = respond_to_batch(gpt2_model, prompt_pt)
            summary = gpt2_tokenizer.decode(sum_pt[0,:])
            print(f'Summary: {summary}')
            
            eval_start_s = 0
            with torch.no_grad():
                pos_reward = s_eval.evaluate(summary)
                
                mnli_input = rand_facts + '</s>' + summary
                mnli_e = mnli_tokenizer(
                    mnli_input, return_tensors='pt', 
                    truncation=True)
                mnli_pred = torch.softmax(
                    mnli_model(**mnli_e).logits, 
                    dim=1).tolist()[0]
                mnli_reward = mnli_pred[2]
                
                reward = 0.2 * pos_reward + 0.8 * mnli_reward
                reward_pt = torch.tensor([reward])
                
            ppo_trainer.step(prompt_pt, sum_pt, reward_pt)
            esc_summary = '"' + escape_text(summary) + '"'
            t_to_q[esc_summary] = pos_reward
            eval_total_s += time.time() - eval_start_s
            
        else:
            t_to_q['None'] = -10
        
        total_s = time.time() - start_s
        if total_s > timeout_s:
            break
    
    p_stats = {}
    p_stats['time'] = total_s
    p_stats['evaluation_time'] = eval_total_s
    
    return t_to_q, p_stats


def vizier_sums(
        timeout_s, connection, table, dim_cols, 
        agg_cols, cmp_pred, nr_facts, nr_preds, preamble, 
        dims_tmp, aggs_txt, all_preds, **kwargs):
    """ Generates and evaluates one random summary.
    
    Args:
        timeout_s: max. time in seconds
        connection: connects to database
        table: summarize this table
        dim_cols: dimension columns
        agg_cols: aggregation columns
        cmp_pred: identifies entity to advertise
        nr_facts: max. number of facts in summary
        nr_preds: max. number of predicates per fact
        preamble: start summary with this text
        dims_tmp: text templates for dimensions
        aggs_txt: text describing aggregates
        all_preds: all possible predicates
    
    Returns:
        Dictionary mapping summaries to reward, statistics
    """
    start_s = time.time()
    cache = nminer.cache.static.EmptyCache()
    q_engine = nminer.sql.query.QueryEngine(
        connection, table, cmp_pred, cache)
    s_gen = nminer.text.sum.SumGenerator(
        all_preds, preamble, dim_cols, dims_tmp, 
        agg_cols, aggs_txt, q_engine)
    s_eval = nminer.text.sum.SumEvaluator()

    pred_cnt = len(all_preds)
    agg_cnt = len(agg_cols)
    text_to_quality = {}
    v_sum = nminer.algs.vizier.VizierSum(nr_facts, nr_preds, pred_cnt, agg_cnt)
    
    counter = 0
    while True:
        counter += 1
        
        trial_name, facts = v_sum.suggest_summary()
        text, _ = s_gen.generate(facts)
        quality = s_eval.evaluate(text)
        text_to_quality[text] = quality
        v_sum.register_feedback(trial_name, quality)
        
        total_s = time.time() - start_s
        if total_s > timeout_s:
            break
    
    stats = {'time':total_s}
    stats.update(cache.statistics())
    stats.update(q_engine.statistics())
    stats.update(s_gen.statistics())
    stats.update(s_eval.statistics())
    
    return text_to_quality, stats


'''
Created on May 3, 2022

@author: immanueltrummer
'''
import argparse
import numpy as np
import pandas as pd
from statsmodels.stats import inter_rater


if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument('survey', type=str, help='Path to all survey responses')
    args = parser.parse_args()
    
    df = pd.read_csv(args.survey)
    a = df.iloc[:,8:12].dropna().apply(lambda r:r.argmax(), axis=1).values
    c = inter_rater.aggregate_raters(np.expand_dims(a, 0))[0]
    print(inter_rater.fleiss_kappa(c, method='unif'))


'''
Created on Aug 8, 2021

@author: immanueltrummer
'''
def cardinality(connection, table):
    """ Returns cardinality estimate. 
    
    Args:
        connection: connection to database
        table: estimate rows in this table
    
    Returns:
        cardinality estimate of given table
    """
    sql = f'select * from {table}'
    cardinality, _ = estimates(connection, sql)
    return cardinality

def estimates(connection, sql):
    """ Generates cost and cardinality estimates.
    
    Args:
        connection: connection to database
        sql: an sql command
    
    Returns:
        tuple with output cardinality and cost estimates
    """
    explain_sql = f'explain (format json) {sql}'
    with connection.cursor() as cursor:
        cursor.execute(explain_sql)
        result = cursor.fetchall()
        rows = result[0]['QUERY PLAN'][0]['Plan']['Plan Rows']
        cost = result[0]['QUERY PLAN'][0]['Plan']['Total Cost']
        return rows, cost


'''
Created on Aug 4, 2021

@author: immanueltrummer
'''
from collections import defaultdict
from nminer.sql.cost import estimates
from nminer.sql.query import AggQuery, GroupQuery
from nminer.cache.dynamic import DynamicCache
from nminer.text.fact import Fact
import logging

class ProCache(DynamicCache):
    """ Implements pro-active caching for RL approach. """
    
    def __init__(self, connection, table, cmp_pred, 
                 nr_facts, nr_preds, all_preds, pred_graph,
                 agg_cols, agg_graph):
        """ Initializes proactive cache.
        
        Args:
            connection: connection to database
            table: name of source table
            cmp_pred: predicate used for comparisons
            nr_facts: number of facts per summary
            nr_preds: number of predicates per fact
            all_preds: list of predicates
            pred_graph: graph of query predicates
            agg_cols: list of aggregation columns
            agg_graph: graph of query aggregates
        """
        super().__init__(connection, table, cmp_pred)
        self.table = table
        self.cmp_pred = cmp_pred
        self.nr_facts = nr_facts
        self.nr_preds = nr_preds
        self.nr_props = nr_preds + 1
        self.all_preds = all_preds
        self.pred_graph = pred_graph
        self.agg_cols = agg_cols
        self.agg_graph = agg_graph
        self.q_to_r = {}
        self.nr_hits = 0
        self.nr_miss = 0

    def set_cur_facts(self, cur_facts):
        """ Inform cache about currently selected facts.
        
        Args:
            cur_facts: facts representing RL state
        """
        self.cur_facts = cur_facts

    def statistics(self):
        """ Returns number of cache hits and misses. """
        return {'cache_hits':self.nr_hits, 'cache_misses':self.nr_miss}

    def update(self):
        """ Proactively cache results close to current facts. """
        q_probs = [(q, p) for q, p in self._query_probs(1).items()]
        for fact in self.cur_facts:
            query = self._props_query(fact.props)
            if not self.can_answer(query):
                self.nr_miss += 1
                logging.debug(f'Query to expand: {query}')
                u_q_probs = [q_p for q_p in q_probs 
                             if not self.can_answer(q_p[0])]
                logging.debug(f'Uncached query probs: {u_q_probs}')
                r_aggs = self._rank_aggs(u_q_probs)
                logging.debug(f'Ranked aggregates: {r_aggs}')
                r_preds = self._rank_preds(u_q_probs, query)
                logging.debug(f'Ranked predicates: {r_preds}')
                exp_g_q = self._expand(query, u_q_probs, 1.5)
                logging.debug(f'Expanded query: {exp_g_q}')
                self.cache(exp_g_q)
            else:
                self.nr_hits += 1
    
    def _coverage(self, g_query, q_probs):
        """ Probability sum of queries covered by group-by query.
        
        Args:
            g_query: a group-by query covering multiple simple queries
            q_probs: list of pairs (queries with associated probabilities)
        
        Returns:
            sum of probability over all covered queries
        """
        agg_prob = 0
        for q, p in q_probs:
            if g_query.contains(q):
                agg_prob += p
        
        return agg_prob
    
    def _expand(self, query, q_probs, max_cost):
        """ Expands given query under cost constraint.
        
        Args:
            query: expand this query
            q_probs: other queries to cover with probabilities
            max_const: maximal relative execution cost
        
        Returns:
            group-by query covering multiple simple queries
        """
        r_aggs = self._rank_aggs(q_probs)
        logging.debug(f'Ranked aggregates: {r_aggs}')
        r_preds = self._rank_preds(q_probs, query)
        logging.debug(f'Ranked predicates: {r_preds}')
        
        g_query = GroupQuery.from_query(query)
        g_sql = g_query.sql()
        _, base_cost = estimates(self.connection, g_sql)
        
        max_cover = 0
        max_exp = g_query
        nr_aggs = len(r_aggs)
        for aggs_ctr in range(nr_aggs):
            aggs = r_aggs[0:aggs_ctr+1]
            e_query = GroupQuery.from_query(query)
            e_query.aggs = aggs
            
            for p in r_preds:
                e_query.preds.add(p)
                e_sql = e_query.sql()
                _, e_cost = estimates(self.connection, e_sql)
                if e_cost <= max_cost * base_cost:
                    cover = self._coverage(e_query, q_probs)
                    if cover > max_cover:
                        max_cover = cover
                        max_exp = e_query
        
        logging.debug(f'Max cover {max_cover} via {max_exp}')
        return max_exp

    def _neighbors(self, s_props):
        """ Generates neighbors in search graph for given properties.
        
        Args:
            s_props: start node in search graph (properties describing fact)
        
        Returns:
            set of neighbors in search graph
        """
        neighbors = []
        for prop in range(self.nr_props):
            
            val = s_props[prop]
            if prop == self.nr_preds:
                n_vals = self.agg_graph.get_reachable(val, 1)
            else:
                n_vals = self.pred_graph.get_reachable(val, 1)
            
            for n_val in n_vals:
                n_props = [p for p in s_props]
                n_props[prop] = n_val
                n_props = tuple(n_props)
                neighbors.append(n_props)
        
        return neighbors
    
    def _props_query(self, props):
        """ Generate aggregation query for property tuple. 
        
        Args:
            props: tuple of properties describing fact
            
        Returns:
            aggregation query associated with fact properties
        """
        fact = Fact.from_props(list(props))
        return AggQuery.from_fact(
            self.table, self.all_preds, 
            self.cmp_pred, self.agg_cols, fact)

    def _query_probs(self, max_steps):
        """ Calculates probability that facts become relevant in few steps.
        
        Args:
            max_steps: consider at most that many steps
        
        Returns:
            Dictionary mapping queries to probabilities
        """
        fact_prob = 1.0/self.nr_facts
        for fact in self.cur_facts:
            p_by_step = []
            p_by_step += [{tuple(fact.props):1}]
            for _ in range(max_steps):
                
                p_last = p_by_step[-1]
                p_next = defaultdict(lambda:0)
                p_by_step.append(p_next)
                
                for s_props, s_prob in p_last.items():                    
                    neighbors = self._neighbors(s_props)
                    logging.debug(f'Neighbors of {s_props}: {neighbors}')
                    
                    # Probability of start state, fact selection, and end state
                    n_prob = 1.0/len(neighbors)
                    t_prob = s_prob * fact_prob * n_prob
                    for n_probs in neighbors:
                        p_next[n_probs] += t_prob
        
        agg_probs = defaultdict(lambda:0)
        for step_probs in p_by_step:
            for props, prob in step_probs.items():
                query = self._props_query(props)
                agg_probs[query] += prob

        return agg_probs

    def _rank_aggs(self, q_probs):
        """ Rank aggregates by their probability.
        
        Args:
            q_probs: pairs of queries and probabilities
            
        Returns:
            list of aggregates ordered by probability (descending)
        """
        p_agg = defaultdict(lambda:0)
        for q, p in q_probs:
            agg = q.agg_col
            p_agg[agg] += p
            
        agg_p_sorted = sorted(p_agg.items(), key=lambda i:i[1], reverse=True)
        return [a for a, _ in agg_p_sorted]
    
    def _rank_preds(self, q_probs, query):
        """ Rank query-compatible predicates based on probabilities.
        
        Args:
            q_probs: pairs of queries and probabilities
            query: rank predicates with same signature as in query
        
        Returns:
            list of compatible predicate groups by probability (descending)
        """
        q_dims = query.pred_cols()
        p_pred = defaultdict(lambda:0)
        for q, p in q_probs:
            p_dims = q.pred_cols()
            if p_dims == q_dims:
                preds = frozenset(q.eq_preds)
                p_pred[preds] += p
        
        pred_p_sorted = sorted(p_pred.items(), key=lambda i:i[1], reverse=True)
        return [p for p, _ in pred_p_sorted]


'''
Created on May 4, 2022

@author: immanueltrummer
'''
import json
import os
import pandas as pd
import pathlib
import psycopg2.extras
import stable_baselines3
import streamlit as st
import sys
import time

cur_file_dir = os.path.dirname(__file__)
src_dir = pathlib.Path(cur_file_dir).parent.parent
sys.path.append(str(src_dir))
print(sys.path)

import nminer.algs.rl
import nminer.sql.pred
import nminer.text.sum

st.set_page_config(page_title='NaturalMiner')
st.markdown('''
NaturalMiner mines large data sets for patterns described in natural language.
''')

root_dir = src_dir.parent
scenarios_path = root_dir.joinpath('demo').joinpath('scenarios.json')
with open(scenarios_path) as file:
    scenarios = json.load(file)
nr_scenarios = len(scenarios)
print('Example scenarios loaded')

selected = st.selectbox(
    'Select Example (Optional)', options=range(nr_scenarios), 
    format_func=lambda idx:scenarios[idx]['scenario'])

with st.expander('Data Source'):
    connection_info = st.text_input(
        'Database connection details (format: "<Database>:<User>:<Password>"):',
        value=scenarios[selected]['dbconnection']).split(':')
    db_name = connection_info[0]
    db_user = connection_info[1]
    db_pwd = connection_info[2] if len(connection_info) > 1 else ''
    table = st.text_input(
        'Name of database table to analyze:', max_chars=100,
        value=scenarios[selected]['table'])

with st.expander('Text Templates'):
    preamble = st.text_input(
        'Preamble text for each fact:', 
        value=scenarios[selected]['preamble'])
    dims_info = st.text_area(
        'One dimension column per line ' +\
        '(format: "<Column>:<TemplateText>"; "<V>" as placeholder):', 
        value=scenarios[selected]['dimensions'], height=20).split('\n')
    aggs_info = st.text_area(
        'One aggregation column per line ' +\
        '(format: "<Column>:<TemplateText>"):', 
        value=scenarios[selected]['aggregates'], height=20).split('\n')
    nr_facts = st.slider(
        'Number of facts:', min_value=1, max_value=3,  
        value=scenarios[selected]['nr_facts'])
    nr_preds = st.slider(
        'Predicates per fact:', min_value=0, max_value=3, 
        value=scenarios[selected]['nr_preds'])

cmp_preds = st.text_area(
    'Data subsets to analyze (one SQL predicate per line):',
    value=scenarios[selected]['predicates'], height=20).split('\n')

label = st.text_input(
    'Pattern (natural language):', max_chars=100, 
    value=scenarios[selected]['goal'])

nr_iterations = st.slider(
    'Number of iterations:', 
    min_value=1, max_value=500, value=200)

print('Generated input elements')


if st.button('Find Pattern!'):
    print('Searching for pattern ...')
    
    result_cols = ['Predicate', 'Facts', 'Quality']
    result_df = pd.DataFrame(columns=result_cols, index=range(0))
    result_table = st.table(result_df)
    
    for cmp_pred in cmp_preds:
        st.write(f'Analyzing data satisfying predicate "{cmp_pred}" ...')
        dims_col_text = [d.split(':') for d in dims_info]
        aggs_col_text = [a.split(':') for a in aggs_info]
        t = {
            'table':table, 
            'dim_cols':[d[0] for d in dims_col_text],
            'agg_cols':[a[0] for a in aggs_col_text],
            'cmp_preds':[cmp_pred],
            'nr_facts':nr_facts, 'nr_preds':nr_preds, 
            'degree':5, 'max_steps':nr_iterations,
            'preamble':preamble, 
            'dims_tmp':[d[1] for d in dims_col_text],
            'aggs_txt':[a[1] for a in aggs_col_text]
        }
        with psycopg2.connect(
            database=db_name, user=db_user, 
            cursor_factory=psycopg2.extras.RealDictCursor) as connection:
            connection.autocommit = True
            
            start_s = time.time()
            all_preds = nminer.sql.pred.all_preds(
                connection, t['table'], 
                t['dim_cols'], cmp_pred)
            sum_eval = nminer.text.sum.SumEvaluator(
                1, 'facebook/bart-large-mnli', label)
            env = nminer.algs.rl.PickingEnv(
                connection, **t, all_preds=all_preds,
                c_type='proactive', cluster=True,
                sum_eval=sum_eval)
            model = stable_baselines3.A2C(
                'MlpPolicy', env, verbose=True, 
                gamma=1.0, normalize_advantage=True)
            model.learn(total_timesteps=nr_iterations)
            total_s = time.time() - start_s
            rated_sums = env.s_eval.text_to_reward
            actual_sums = [i for i in rated_sums.items() if i[0] is not None]
            sorted_sums = sorted(actual_sums, key=lambda s: s[1])
            if sorted_sums:
                b_sum = sorted_sums[-1]
            else:
                b_sum = ('(No valid summary generated)', -10)
            
            new_row = pd.DataFrame([
                [cmp_pred, b_sum[0], b_sum[1]]], 
                columns=result_cols)
            result_table.add_rows(new_row)
            
    st.write('All data subsets generated!')


'''
Created on Jul 31, 2021

@author: immanueltrummer
'''
from nminer.cache.common import AggCache
from nminer.sql.pred import is_pred, pred_sql
import logging
import time
from psycopg2._psycopg import QueryCanceledError

class EmptyCache(AggCache):
    """ Dummy cache that remains empty. """
    
    def __init__(self):
        self.nr_miss = 0
    
    def can_answer(self, _):
        self.nr_miss += 1
        return False
    
    def get_result(self, _):
        raise NotImplementedError
    
    def statistics(self):
        return {'cache_hits':0, 'cache_misses':self.nr_miss}
    
    def update(self):
        pass


class CubeCache(AggCache):
    """ Answers all queries from initially created data cube. """
    
    def __init__(self, connection, src_table, dim_cols, 
                 cmp_pred, agg_cols, timeout_s):
        """ Initializes data cube used for all lookups.
        
        Args:
            connection: database connection
            src_table: build cube from this table
            dim_cols: dimension columns of cube
            cmp_pred: predicate for comparisons
            agg_cols: aggregation columns of cube
            timeout_s: timeout in seconds
        """
        self.connection = connection
        self.src_tbl = src_table
        self.dim_cols = dim_cols
        self.cmp_pred = cmp_pred
        self.agg_cols = agg_cols
        self.timeout_s = timeout_s
        self.cube_tbl = 'cpcube'
        self.nr_hits = 0
        self.nr_miss = 0
        self._clear()
        self.have_cube = self._create_cube()
    
    def can_answer(self, _):
        if self.have_cube:
            self.nr_hits += 1
            return True
        else:
            self.nr_miss += 1
            return False
    
    def get_result(self, query):
        """ Get result for query from cached cube.
        
        Args:
            query: aggregation query
            
        Returns:
            query result
        """
        a = query.agg_col
        q_parts = [f'select case when cmp_c = 0 or s_{a} = 0 then NULL']
        q_parts += [f'else (cmp_s_{a}/cmp_c)/(s_{a}/c) end as rel_avg,']
        q_parts += [f'cmp_c as row_count from {self.cube_tbl}']
        
        w_parts = [pred_sql(c, v) for c, v in query.eq_preds if is_pred((c, v))]
        p_cols = query.pred_cols()
        for dim in self.dim_cols:
            if dim not in p_cols:
                w_parts += [f'{dim} is NULL']
        w_clause = ' AND '.join(w_parts)
        
        sql = ' '.join(q_parts) + ' where ' + w_clause
        with self.connection.cursor() as cursor:
            logging.debug(f'Cache lookup via "{sql}"')
            cursor.execute(sql)
            if cursor.rowcount == 0:
                return None
            else:
                r = cursor.fetchone()
                return r['rel_avg'], r['row_count']

    def statistics(self):
        """ Returns number of cache hits and misses. """
        return {'cache_hits':self.nr_hits, 'cache_misses':self.nr_miss}

    def update(self):
        pass
    
    def _clear(self):
        """ Clears cache content. """
        with self.connection.cursor() as cursor:
            cursor.execute(f'drop table if exists {self.cube_tbl}')
    
    def _create_cube(self):
        """ Creates cube used for answering all queries.
        
        Returns:
            true iff cube was created until timeout
        """
        if len(self.dim_cols) > 12:
            return False
        
        s_parts = ['select count(*) as c']
        s_parts += [f'sum(case when {self.cmp_pred} then 1 else 0 end) as cmp_c']
        s_parts += self.dim_cols
        
        for agg_col in self.agg_cols:
            s_parts += [f'sum({agg_col}) as s_{agg_col}']
            s_parts += [f'sum(case when {self.cmp_pred} then {agg_col}' \
                        f' else 0 end) as cmp_s_{agg_col}']
            
        sql = f'create unlogged table {self.cube_tbl} as (' + \
            ', '.join(s_parts) + f' from {self.src_tbl} ' + \
            ' group by cube (' + ', '.join(self.dim_cols) + '))'
        logging.debug(f'About to create cube via "{sql}"')

        start_s = time.time()
        success = False
        with self.connection.cursor() as cursor:
            cursor.execute(f"set statement_timeout = '{self.timeout_s}s'")
            try:
                cursor.execute(sql)
                success = True
            except QueryCanceledError:
                logging.debug('Timeout while creating cube.')
                success = False
            finally:
                cursor.execute('set statement_timeout = 0')
        total_s = time.time() - start_s
        logging.debug(f'Created cube in {total_s} seconds; success: {success}')
            
        return success


'''
Created on Apr 27, 2022

@author: immanueltrummer
'''
import nminer.text.fact
import datetime
import googleapiclient.discovery
import google.cloud.storage
import os
import time


class VizierSum():
    """ Baseline using Google's Vizier to generate summaries. """
    
    def __init__(self, nr_facts, nr_preds, pred_cnt, agg_cnt):
        """ Initializes summary generation for specific dimensions.
        
        Args:
            nr_facts: number of facts in summaries
            nr_preds: maximal number of predicates per fact
            pred_cnt: number of predicate options
            agg_cnt: number of aggregate options
        """
        self.nr_facts = nr_facts
        self.nr_preds = nr_preds
        self.pred_cnt = pred_cnt
        self.agg_cnt = agg_cnt
        
        self.project_id = os.environ['VIZIER_PROJECT_ID']
        timestamp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        self.study_id = f'itrummer_study_{timestamp}'
        self.study_parent = f'projects/{self.project_id}/locations/us-central1'
        self.study_name = f'{self.study_parent}/studies/{self.study_id}'
        self.client_id = 'summary-client'

        self.client = self._client()
        self.operations = self.client.projects().locations().operations()
        self.studies = self.client.projects().locations().studies()
        self.trials = self.studies.trials()
        study_config = self._study_config(nr_facts, nr_preds, pred_cnt, agg_cnt)
        self._create_study(study_config)
    
    def register_feedback(self, trial_name, quality):
        """ Provide feedback on suggested parameter values. 
        
        Args:
            trial_name: name of evaluated trial
            quality: quality of evaluated summary
        """
        metric = {'metric': 'quality', 'value': quality}
        measurement = {'step_count': 1, 'metrics': [metric,]}
        self.trials.addMeasurement(
            name=trial_name, 
            body={'measurement': measurement}).execute()
        self.trials.complete(name=trial_name).execute()

    def suggest_summary(self):
        """ Generate suggestions for summaries to evaluate. 
        
        Returns:
            trial name and associated summary
        """
        body = {'client_id': self.client_id, 'suggestion_count': 1}
        trial_response = self.trials.suggest(
            parent=self.study_name, body=body).execute()
        op_id = trial_response['name'].split('/')[-1]
        op_name = f'{self.study_parent}/operations/{op_id}'
        
        get_op = self.operations.get(name=op_name)
        while not get_op.execute().get('done', False):
            time.sleep(1)
        
        trial_info = get_op.execute()['response']['trials'][0]
        trial_id = int(trial_info['name'].split('/')[-1])
        trial_name = f'{self.study_parent}/studies/{self.study_id}/trials/{trial_id}'
        trial = self.trials.get(name=trial_name).execute()
        
        return trial_name, self._extract_summary(trial)
    
    def _client(self):
        """ Generates client to access Vizier. 
        
        Returns:
            client for accessing ML API
        """
        client = google.cloud.storage.Client(self.project_id)
        bucket = client.get_bucket('caip-optimizer-public')
        blob = bucket.get_blob('api/ml_public_google_rest_v1.json')
        api_doc = blob.download_as_string()
        return googleapiclient.discovery.build_from_document(api_doc)

    def _create_study(self, study_config):
        """ Generates a new study. 
        
        Args:
            study_config: configuration of study to create
        """
        request = self.studies.create(
            parent=self.study_parent, 
            studyId=self.study_id, 
            body=study_config)
        request.execute()
    
    def _extract_summary(self, trial):
        """ Extract summary from trial. 
        
        Args:
            trial: suggested parameter values, describing summary
        
        Returns:
            summary specified by trial suggestions
        """
        summary = []
        for _ in range(self.nr_facts):
            fact = nminer.text.fact.Fact(self.nr_preds)
            summary.append(fact)
        
        for p in trial['parameters']:
            name = p['parameter']
            properties = name.split('-')
            fact_idx = int(properties[1])
            fact = summary[fact_idx]
            
            value = int(p['intValue'])
            p_type = properties[0]
            if p_type == 'A':
                fact.set_agg(value)
            elif p_type == 'P':
                pred_idx = int(properties[2])
                fact.set_pred(pred_idx, value)
            else:
                raise ValueError(f'Unsupported parameter type: {p_type}')
        
        return summary
    
    def _study_config(self, nr_facts, nr_preds, pred_cnt, agg_cnt):
        """ Generates configuration for study.
        
        Args:
            nr_facts: number of facts in summary
            nr_preds: number of possible predicates
            pred_cnt: number of predicate options
            agg_cnt: number of aggregate options
        
        Returns:
            a study configuration (as Python dictionary)
        """
        study = {'algorithm': 'ALGORITHM_UNSPECIFIED'}
        parameters = []
        study['parameters'] = parameters
        for fact_idx in range(nr_facts):
            
            p = {}
            p['parameter'] = f'A-{fact_idx}'
            p['type'] = 'INTEGER'
            int_value_spec = {}
            int_value_spec['min_value'] = 0
            int_value_spec['max_value'] = agg_cnt - 1
            p['integer_value_spec'] = int_value_spec
            parameters.append(p)
            
            for pred_idx in range(nr_preds):
                p = {}
                p['parameter'] = f'P-{fact_idx}-{pred_idx}'
                p['type'] = 'INTEGER'
                int_value_spec = {}
                int_value_spec['min_value'] = 0
                int_value_spec['max_value'] = pred_cnt - 1
                p['integer_value_spec'] = int_value_spec
                parameters.append(p)

        metric = {}
        study['metrics'] = [metric]
        metric['metric'] = 'quality'
        metric['goal'] = 'MAXIMIZE'
        
        return {'study_config': study}


'''
Created on Sep 4, 2021

@author: immanueltrummer
'''
import argparse
import pandas as pd

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument('in_path', type=str, help='path to input file')
    parser.add_argument('out_path', type=str, help='path to output file')
    args = parser.parse_args()
    
    df = pd.read_csv(args.in_path, sep=',')
    df.drop('text', axis=1, inplace=True)
    df.drop('cluster', axis=1, inplace=True)
    
    batch_df = df.query('approach in ["simple", "cluster"]')
    groups = batch_df.groupby(['nrfacts', 'nrpreds', 'pred'])
    aggs = groups.agg('max').reset_index()
    aggs['approach'] = 'bestof'
    
    all_df = pd.concat([df, aggs], ignore_index=True)
    all_df.to_csv(args.out_path, sep=',')


'''
Created on Aug 15, 2021

@author: immanueltrummer
'''
import collections
import nminer.algs.rl
import nminer.algs.sample
import nminer.cache.multi
import nminer.sql.query
import nminer.text.fact
import dataclasses
import gym
from gym import spaces
import logging
import numpy as np
import random
from sklearn.cluster import KMeans
from stable_baselines3 import A2C
import statistics
from nminer.sql.pred import pred_sql


@dataclasses.dataclass(frozen=True)
class EvaluationResult():
    summary: str
    quality: float


def eval_solution(connection, batch, all_preds, solution):
    """ Evaluates solution to batch summarization problem.
    
    Careful: this function assumes that comparison predicates
    are unary equality predicates on the comparison column.
    If this is not the case, it adds unnecessary overheads
    (the result is still correct). Might refine later.
    
    Args:
        connection: connection to database
        batch: description of test case
        all_preds: all predicates on dimension columns
        solution: maps each predicate to a summary template
    
    Returns:
        dictionary mapping predicates to summaries with reward
    """
    table = batch['general']['table']
    preamble = batch['general']['preamble']
    dim_cols = batch['general']['dim_cols']
    dims_tmp = batch['general']['dims_tmp']
    agg_cols = batch['general']['agg_cols']
    aggs_txt = batch['general']['aggs_txt']
    cmp_col = batch['cmp_col']
    
    # collect items with same summary template
    sum_to_preds = collections.defaultdict(lambda:[])
    for pred, sum_tmp in solution.items():
        sum_to_preds[sum_tmp] += [pred]
    
    result = {}
    for sum_tmp, cmp_preds in sum_to_preds.items():
        cache = nminer.cache.multi.MultiItemCache(
            connection, table, cmp_col, 
            all_preds, agg_cols, sum_tmp)
        s_eval = nminer.text.sum.SumEvaluator()
        
        logging.info('Generating text summaries ...')
        sums = []
        for cmp_pred in cmp_preds:
            q_engine = nminer.sql.query.QueryEngine(
                connection, table, cmp_pred, cache)
            s_gen = nminer.text.sum.SumGenerator(
                all_preds, preamble, dim_cols,
                dims_tmp, agg_cols, aggs_txt,
                q_engine)
            sum_tmp = solution[cmp_pred]
            sum_facts = [nminer.text.fact.Fact.from_props(p) for p in sum_tmp]
            d_sum, _ = s_gen.generate(sum_facts)
            sums += [d_sum]

        logging.info('Evaluating summary quality ...')
        scores = s_eval.evaluate_batch(sums)
        for cmp_pred, d_sum, score in zip(cmp_preds, sums, scores):
            result[cmp_pred] = EvaluationResult(d_sum, score)
    
    return result


def simple_batch(connection, batch, all_preds):
    """ Simple baseline using same summary template for entire batch.
    
    Args:
        connection: connection to database
        batch: a batch of summarization tasks
        all_preds: all predicates on dimensions
    
    Returns:
        Dictionary mapping each predicate to summary template
    """
    test_case = batch['general'].copy()
    all_cmp_preds = batch['predicates']
    if not all_cmp_preds:
        return {}

    to_select = min(3, len(all_cmp_preds))
    cmp_preds = random.choices(all_cmp_preds, k=to_select)
    test_case['cmp_preds'] = cmp_preds

    env = nminer.algs.rl.PickingEnv(
        connection, **test_case, all_preds=all_preds,
        c_type='proactive', cluster=True)
    model = A2C(
        'MlpPolicy', env, verbose=True, 
        gamma=1.0, normalize_advantage=True)
    model.learn(total_timesteps=200)
    
    if env.props_to_rewards:
        best = sorted(
            env.props_to_rewards.items(), 
            key=lambda i:i[1])[-1]
        best_props = best[0]
    else:
        nr_facts = test_case['nr_facts']
        nr_preds = test_case['nr_preds']
        fact = nminer.text.fact.Fact(nr_preds)
        best_props = nr_facts * [fact]
    
    return {p:best_props for p in batch['predicates']}


class IterativeClusters():
    """ Iteratively divides items into clusters to generate summaries. """
    
    def __init__(self, connection, batch, dim_preds):
        """ Initialize for given problem.
        
        Args:
            connection: connection to database
            batch: batch containing all items
            dim_preds: all predicates on dimensions
        """
        self.connection = connection
        self.cmp_preds = batch['predicates']
        self.dim_preds = dim_preds
        empty_eval = EvaluationResult(None, -10)
        cmp_to_eval = {cmp:empty_eval for cmp in self.cmp_preds}
        self.clusters = {0:(batch, cmp_to_eval)}

    def iterate(self):
        """ Performs one iteration. """
        logging.info('Splitting cluster ...')
        to_split_idx = self._select()
        to_split = self.clusters[to_split_idx]
        split = self._even_split(to_split)
        del self.clusters[to_split_idx]
        
        logging.info('Summarizing splits ...')
        for cluster in split:
            sketch = self._pick_sketch(cluster)
            batch, prior_eval = cluster
            cmp_preds = batch['predicates']
            cluster_size = len(cmp_preds)
            logging.info(f'Summarizing cluster of size {cluster_size} ...')
            new_eval = eval_solution(
                self.connection, batch, 
                self.dim_preds, 
                {p:sketch for p in cmp_preds})
            
            cluster_id = self._next_ID()
            combined_eval = self._prune_summaries(prior_eval, new_eval)
            self.clusters[cluster_id] = batch, combined_eval
        
        self._log_statistics()

    def _even_split(self, cluster):
        """ Splits items, sorted by text quality, evenly into two halves.
        
        Args:
            cluster: batch with associated evaluation
        
        Returns:
            list containing two clusters resulting from the split
        """
        batch, eval_s = cluster
        nr_preds = len(eval_s)
        preds_by_qual = sorted(eval_s.keys(), key=lambda p:eval_s[p].quality)

        split_clusters = []
        middle = round(nr_preds/2)
        for s in [slice(0, middle), slice(middle, nr_preds)]:
            cmp_preds = preds_by_qual[s]
            split_batch = batch.copy()
            split_batch['predicates'] = cmp_preds
            split_eval = {p:eval_s[p] for p in cmp_preds}
            split_cluster = split_batch, split_eval
            split_clusters.append(split_cluster)
        
        return split_clusters

    def _log_statistics(self):
        """ Logs statistics on quality of generated summaries. """
        all_scores = []
        for _, evaluations in self.clusters.values():
            all_scores += [e.quality for e in evaluations.values()]
        avg_score = statistics.mean(all_scores)
        logging.info(f'Avg. summary quality: {avg_score}')

    def _next_ID(self):
        """ Returns next available batch ID.
        
        Returns:
            integer representing next batch ID
        """
        ids = self.clusters.keys()
        return 1+max(ids) if ids else 0

    def _pick_sketch(self, cluster):
        """ Simple baseline using same summary template for entire batch.
        
        Args:
            cluster: tuple of batch and prior results
        
        Returns:
            best summary sketch
        """
        batch, s_eval = cluster
        batch_cmp_preds = batch['predicates']

        to_select = min(5, len(batch_cmp_preds))
        cmp_preds = random.choices(batch_cmp_preds, k=to_select)
        test_case = batch['general'].copy()
        test_case['cmp_preds'] = cmp_preds
        
        prior_results = {p:e for p, e in s_eval.items() if p in cmp_preds}
        logging.info(f'Best prior results: {prior_results}')
        prior_quality = {p:e.quality for p, e in prior_results.items()}
        env = nminer.algs.rl.PickingEnv(
            self.connection, **test_case, 
            all_preds=self.dim_preds,
            c_type='proactive', cluster=True,
            prior_best=prior_quality)
        model = A2C(
            'MlpPolicy', env, verbose=True, 
            gamma=1.0, normalize_advantage=True)
        model.learn(total_timesteps=50)
        
        if env.props_to_rewards:
            best = sorted(
                env.props_to_rewards.items(), 
                key=lambda i:i[1])[-1]
            logging.info(f'Best sketch: {best}')
            best_props = best[0]
        else:
            logging.info('No valid solution produced')
            nr_facts = test_case['nr_facts']
            nr_preds = test_case['nr_preds']
            fact = nminer.text.fact.Fact(nr_preds)
            best_props = nr_facts * [fact]
        
        return best_props

    def _priority_avg(self, cluster):
        """ Evaluates priority for splitting cluster using average quality.
        
        Args:
            cluster: tuple of batch and associated evaluations
        
        Returns:
            splitting priority (high priority means likely split)
        """
        _, s_eval = cluster
        # bad_items = [i for i, (_, q) in s_eval.items() if q < 0]
        # return len(bad_items)
        return - statistics.mean([v.quality for v in s_eval.values()])
    
    def _priority_count(self, cluster):
        """ Evaluates priority for splitting batch using count of bad items.
        
        Args:
            cluster: batch with associated evaluation
        
        Returns:
            splitting priority (high priority means likely split)
        """
        _, s_eval = cluster
        bad_items = [i for i, e in s_eval.items() if e.quality < 0]
        return len(bad_items)

    def _prune_summaries(self, eval_1, eval_2):
        """ Prune evaluated summaries and keep only the best. 
        
        Args:
            eval_1: dictionary mapping predicates to evaluated summaries
            eval_2: second dictionary mapping predicates to evaluations
        
        Returns:
            dictionary mapping predicates to best found evaluation
        """
        pruned = {}
        for cmp, e_1 in eval_1.items():
            e_2 = eval_2[cmp]
            if e_1.quality >= e_2.quality:
                pruned[cmp] = e_1
            else:
                pruned[cmp] = e_2
        return pruned

    def _select(self):
        """ Select one of current batches to split. 
        
        Returns:
            index of batch to split
        """
        cluster_ids = self.clusters.keys()
        return max(cluster_ids, 
                   key=lambda c_id:
                   self._priority_avg(
                       self.clusters[c_id]))
    
    def _signum_split(self, b_e):
        """ Splits items based on the signum of text quality.
        
        Args:
            b_e: batch with associated evaluation
        
        Returns:
            list of two batches splitting the input batch
        """
        batch, eval_s = b_e
        preds_1 = [p for p, e in eval_s.items() if e.quality < 0]
        preds_2 = [p for p, e in eval_s.items() if e.quality >= 0]
        
        split_batches = []
        for preds in [preds_1, preds_2]:
            split_batch = batch.copy()
            split_batch['predicates'] = preds
            split_batches.append(split_batch)
        
        return split_batches


class ClusterEnv(gym.Env):
    """ Environment teaching agent how to cluster items for summarization. """
    
    def __init__(self, connection, batch, all_preds, 
                 cmp_preds, nr_features, raw_features):
        """ Initializes weighting for given number of features.
        
        Args:
            connection: connection to database
            batch: batch of summarization tasks
            all_preds: all possible predicates on dimensions
            cmp_preds: predicates identifying items for comparison
            nr_features: number of features to weigh
            raw_features: cluster based on these features
        """
        self.connection = connection
        self.batch = batch
        self.all_preds = all_preds
        self.cmp_preds = cmp_preds
        self.nr_features = nr_features
        self.raw_features = raw_features
        self.nr_items = len(cmp_preds)
        self.action_space = spaces.Box(
            low=0, high=1, shape=(nr_features,), 
            dtype=np.float32)
        self.observation_space = spaces.Discrete(1)
        self.best_clusters = None
        self.best_reward = float('-inf')
    
    def reset(self):
        return 0
        
    def step(self, action):
        """ Executes one step of reinforcement learning. 
        
        Args:
            action: new set of weights to try
        """
        weights = action
        clusters = self._cluster(weights)
        reward = self._eval_clusters(clusters)
        if reward > self.best_reward:
            self.best_clusters = clusters

        logging.debug(f'Reward {reward} for weights {weights}')
        for c_id, c_items in clusters.items():
            logging.debug(f'Cluster {c_id}: {c_items}')
        return 0, reward, False, {}
    
    def _cluster(self, weights):
        """ Clusters items using features weighted by given weights. 
        
        Args:
            weights: multiple distance for each feature by those weights
        
        Returns:
            item clusters (set of item sets)
        """
        w_features = []
        for cmp_pred in self.cmp_preds:
            cur_w_features = []
            cur_r_features = self.raw_features[cmp_pred]
            for f, w in zip(cur_r_features, weights):
                cur_w_features.append(f*w)
            w_features.append(cur_w_features)
        
        X = np.array(w_features)
        kmeans = KMeans(n_clusters=5).fit(X)
        
        clusters = collections.defaultdict(lambda:set())
        for item, label in enumerate(kmeans.labels_):
            clusters[label].add(self.cmp_preds[item])
        
        return clusters
    
    def _eval_clusters(self, clusters):
        """ Estimate average summary quality with given clusters.
        
        Args:
            clusters: maps cluster IDs to sets of comparison predicates

        Returns:
            estimated average quality when generating one summary per cluster
        """
        # select random cluster, probability weighted by cluster size
        item_idx = random.randint(0, self.nr_items-1)
        cmp_pred = self.cmp_preds[item_idx]
        c_id = [c for c, items in clusters.items() if cmp_pred in items][0]
        
        # generate summary and estimate generalization for items in cluster
        cluster_items = list(clusters[c_id])
        cluster_batch = self.batch.copy()
        cluster_batch['predicates'] = random.choices(cluster_items, k=5)
        
        solution = simple_batch(
            self.connection, cluster_batch, 
            self.all_preds)
        eval_sol = eval_solution(
            self.connection, cluster_batch, 
            self.all_preds, solution)

        mean = statistics.mean([r for (_, r) in eval_sol.values()])
        logging.debug(f'Mean {mean} for {eval_sol}')
        return mean

class BatchProcessor():
    """ Generates summaries for item clusters. """
    
    def __init__(self, connection, batch, all_preds):
        """ Initializes for given batch. 
        
        Args:
            connection: connection to database
            batch: batch of test cases
            all_preds: predicates on dimensions
        """
        self.connection = connection
        self.batch = batch
        self.all_preds = all_preds
        self.table = batch['general']['table']
        self.cmp_preds = batch['predicates']
        self.nr_features, self.raw_features = self._get_features()
        self.cluster_env = ClusterEnv(
            connection, batch, all_preds, 
            self.cmp_preds, self.nr_features, 
            self.raw_features)
    
    def summarize(self):
        """ Generates summaries for clusters of items.
        
        Returns:
            dictionary mapping cluster IDs to a solution (item->summary)
        """
        model = A2C(
            'MlpPolicy', self.cluster_env, verbose=True, 
            gamma=1.0, normalize_advantage=True)
        model.learn(total_timesteps=10)
        clusters = self.cluster_env.best_clusters
        logging.info(f'Clusters: {clusters}')
        
        result = {}
        for c_id, cmp_preds in clusters.items():
            logging.info(f'Processing cluster {cmp_preds}')
            cluster_batch = self.batch.copy()
            cluster_batch['predicates'] = list(cmp_preds)
            result[c_id] = simple_batch(
                self.connection, cluster_batch, self.all_preds)
        
        return result
    
    def _get_features(self):
        """ Collect raw features used to cluster items.
        
        Careful: this implementation assumes that comparison
        predicates are unary equality predicates on the same
        column (therefore, all features can be retrieved by
        a single group-by query).
        
        Returns:
            Nr. features, dictionary mapping items to feature vectors
        """
        results = {}
        agg_cols = self.batch['general']['agg_cols']
        dim_cols = self.batch['general']['dim_cols']
        cmp_col = self.batch['cmp_col']
        
        with self.connection.cursor() as cursor:
            
            agg_features = [f'avg({a}) as {a}' for a in agg_cols]
            # for agg_col in agg_cols:
                # for dim_col in dim_cols:
                    # sql = f'select {dim_col} as val from {self.table} limit 1'
                    # cursor.execute(sql)
                    # row = cursor.fetchone()
                    # val = row['val']
                    # e_val = nminer.sql.pred.sql_esc(val)
                    # agg_feature = \
                        # f"(select sum(case when {dim_col} = '{e_val}' " \
                        # f"then {agg_col} else 0 end))/(select count({agg_col}))"
                    # agg_features.append(agg_feature)
            
            nr_features = len(agg_features)
            sql = 'select ' + ', '.join(agg_features) + \
                f', {cmp_col} as cmp_val ' + \
                f'from {self.table} group by {cmp_col}'
            logging.debug(f'Feature query: {sql}')
            cursor.execute(sql)
            rows = cursor.fetchall()
            for row in rows:
                result = [row[a] for a in agg_cols]
                result = [float(r) if r is not None else 0 for r in result]
                cmp_val = row['cmp_val']
                cmp_pred = pred_sql(cmp_col, cmp_val)
                results[cmp_pred] = result
        
        logging.debug(results)
        return nr_features, results


class SubModularIterative():
    """ Iteratively improves summaries based on sub-modularity. """
    
    def __init__(self, connection, batch, all_preds):
        """ Initialize for given problem.
        
        Args:
            connection: connection to database
            batch: batch containing all items
            all_preds: all predicates on dimensions
        """
        self.connection = connection
        self.batch = batch
        self.all_preds = all_preds
        self.all_cmp_preds = batch['predicates']
        empty_result = EvaluationResult('', -1)
        self.best_sums = {p:empty_result for p in self.all_cmp_preds}

    def iterate(self):
        """ Performs one iteration to improve summary quality. 
        
        The algorithm selects a sample of items. Next, it uses
        reinforcement learning to find a summary sketch which
        improves summary quality for those items maximally.
        Each item is assigned to the summary with maximal 
        quality (either the new summary or a prior summary).
        """
        logging.info('Picking new sketch ...')
        best_props = self._pick_sketch()
        solution = {
            cmp_pred:best_props for 
            cmp_pred in self.all_cmp_preds}
        logging.info('Evaluating sketch ...')
        s_evals = eval_solution(
            self.connection, self.batch, 
            self.all_preds, solution)
        logging.info('Pruning sketches ...')
        for cmp_pred, new_eval in s_evals.items():
            prior_eval = self.best_sums[cmp_pred]
            if prior_eval.quality < new_eval.quality:
                self.best_sums[cmp_pred] = new_eval
        self._log_statistics()
    
    def _log_statistics(self):
        """ Log statistics on quality of generated summaries. """
        q_values = [e.quality for e in self.best_sums.values()]
        mean_quality = statistics.mean(q_values)
        logging.info(f'Average quality: {mean_quality}')
    
    def _pick_sketch(self):
        """ Pick a summary sketch that is optimally complementary.
        
        Picks a summary sketch that optimally complements previously
        considered summary sketches. This means that the improvement
        by assigning items not covered well by previous summaries
        to this sketch is maximal.
        
        Returns:
            List of facts representing summary sketch
        """
        print('Selecting next sketch ...')
        test_case = self.batch['general'].copy()
        all_cmp_preds = self.batch['predicates']
        if not all_cmp_preds:
            return {}
    
        to_select = min(5, len(all_cmp_preds))
        cmp_preds = random.choices(all_cmp_preds, k=to_select)
        test_case['cmp_preds'] = cmp_preds
        prior_evals = {p:self.best_sums[p] for p in cmp_preds}
        prior_quality = {p:e.quality for p, e in prior_evals.items()}
        logging.info(f'Prior evaluation results: {prior_evals}')
        
        # table = test_case['table']
        # sample_ratio = 0.01
        # logging.info(f'Sampling {table} with ratio {sample_ratio} ...')
        # sample_table = nminer.create_sample(
            # self.connection, table, sample_ratio)
        # test_case['table'] = sample_table
    
        total_timesteps = 200
        logging.info(f'Iterating for {total_timesteps} steps ...')
        env = nminer.algs.rl.PickingEnv(
            self.connection, **test_case, 
            all_preds=self.all_preds,
            c_type='proactive', cluster=True,
            prior_best=prior_quality)
        model = A2C(
            'MlpPolicy', env, verbose=True, 
            gamma=1.0, normalize_advantage=True)
        model.learn(total_timesteps=total_timesteps)
        
        if env.props_to_rewards:
            best = sorted(env.props_to_rewards.items(), key=lambda i:i[1])[-1]
            logging.info(f'Best summary: {best}')
            best_props = best[0]
        else:
            logging.info('No summary found')
            nr_facts = test_case['nr_facts']
            nr_preds = test_case['nr_preds']
            fact = nminer.text.fact.Fact(nr_preds)
            best_props = nr_facts * [fact]
        
        return best_props


'''
Created on Jul 21, 2021

@author: immanueltrummer
'''
scenarios = [
    {
        'general':{
            'table':'laptops', 
            'dim_cols':['brand', 'processor_type', 'graphics_card', 'disk_space', 'ratings_5max'], 
            'agg_cols':['display_size', 'discount_price', 'old_price'], 
            'nr_facts':1, 'nr_preds':2, 'degree':5, 'max_steps':20,
            'preamble':'Among all laptops', 
            'dims_tmp':['with <V> brand', 'with <V> CPU', 'with <V> graphics card', 
                        'with <V> disk space', 'with <V> stars'],
            'aggs_txt':['its display size', 'its discounted price', 'its old price']
            },
        'cmp_col':'laptop_name',
        'predicates':[
            "laptop_name='Swift 3 SF314-54G-87HB'", 
            "laptop_name='XPS 15 9570'",
            "laptop_name='Swift 1 SF114-32-C4GB'",
            "laptop_name='MacBook Pro (Retina + Touch Bar)'",
            "laptop_name='Pavilion 13-an0001nx'"
            ]
    },
    {
        'general':{
            'table': 'sods19',
            'dim_cols':['satisfaction', 'dependents', 'optimism', 'codereviews', 'unittests', 
                        'languagesused', 'country', 'coarsesalary', 'orgsize', 'jobtype',
                        'experience', 'devrole', 'gender'],
            'agg_cols':['salary'],
            'nr_facts':1, 'nr_preds':2, 'degree':5, 'max_steps':20,
            'preamble':'Among all developers',
            'dims_tmp':['who are <V> with their job', 
                        'who answered "<V>" when asked if they care for dependents',
                        'who answered "<V>" when asked if they feel optimistic',
                        'who answered "<V>" when asked if they do code reviews',
                        'who answered "<V>" when asked if they do unit tests',
                        'who used <V> as programming languages',
                        'from <V>', 'with a salary <V>', 'working for a <V> company',
                        'working <V>', 'who are <V>', 'who work as <V>', 'who are <V>'],
            'aggs_txt':[', the salary']
            },
        'cmp_col':'languagesused',
        'predicates':[
            "languagesused like '%Bash%'",
            "languagesused like '%HTML%'",
            "languagesused like '%Python%'",
            "languagesused like '%C++%'",
            "languagesused like 'C#'"
            ]
    },
    {
        'general':{
            'table':'flights',
            'dim_cols':['flight_date', 'flight_nr', 'start_airport',
                        'destination_airport', 'planned_departure',
                        'actual_departure', 'wheels_off_time',
                        'wheels_on_time', 'planned_arrival_time',
                        'actual_arrival_time', 'cancelled',
                        'cancellation_code', 'diverted',
                        'planned_flight_time', 
                        'distance'],
            'agg_cols':['departure_delay', 'taxi_out_time', 'taxi_in_time',
                        'arrival_delay', 'actual_time', 'air_time',
                        'carrier_delay', 'weather_delay', 'system_delay',
                        'security_delay', 'late_aircraft_delay'],
            'nr_facts':1, 'nr_preds':2, 'degree':5, 'max_steps':20,
            'preamble':'Among all flights',
            'dims_tmp':['on <V>', 'with number <V>', 'from <V>', 
                        'to <V>', 'scheduled for departure at <V>', 
                        'that departed at <V>', 'that took off at <V>',
                        'that landed at <V>', 'scheduled to arrive at <V>',
                        'that arrived at <V>', 'that were <V>',
                        'with cancellation code <V>', 'that were <V>',
                        'with a planned flight time of <V> minutes',
                        'with a flight distance of <V> miles'
                        ],
            'aggs_txt':[', the departure delay', ', the taxi time at departure',
                        ', the taxi time at the destination', 
                        ', the arrival delay', ', the actual flight time',
                        ', the air time', ', the carrier delay',
                        ', weather delay', ', airport operations delay',
                        ', security delay', ', delay to to late aircrafts']
            },
        'cmp_col':'airline',
        'predicates':[
            "airline='NK'", "airline='WN'", "airline='DL'", 
            "airline='AA'", "airline='OH'" 
            ]
    },
    {
        'general':{
            'table':'liquor', 
            'dim_cols':['city', 'countyname', 'categoryname', 
                        'storename', 'vendorname'], 
            'agg_cols':['bottlessold', 'salevalue', 'volumesold'], 
            'nr_facts':1, 'nr_preds':2, 'degree':5, 'max_steps':20, 
            'preamble':'Among all liquors', 
            'dims_tmp':['sold in <V>', 'sold in <V>', 'of type <V>',
                        'sold at <V>', 'produced by <V>'],
            'aggs_txt':[', the number of bottles per sale', 
                        ', the dollar value per sale', ', the volume per sale']
            },
        'cmp_col':'itemname',
        'predicates':[
            "itemname='Johnnie Walker Black'",
            "itemname='Crown Royal Regal Apple Mini'",
            "itemname='Jim Beam Vanilla'",
            "itemname='Smirnoff Vodka 80 Prf'",
            "itemname='Jim Beam'"
            ]
    }
]

def generate_testcases():
    """ Generates test cases for cherry-picking. 
    
    Returns:
        list of batches of test cases on same data
    """
    test_batches = []
    for s in scenarios:
        general = s['general']
        test_cases = []
        test_batches.append(test_cases)
        for cmp_pred in s['predicates']:
            test_case = general.copy()
            test_case['cmp_pred'] = cmp_pred
            test_cases.append(test_case)
        
    return test_batches


'''
Created on Aug 15, 2021

@author: immanueltrummer
'''
import argparse
import nminer.eval.bench
import nminer.sql.pred
import json
import psycopg2


if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument('db', type=str, help='Name of database to access')
    parser.add_argument('user', type=str, help='Database user name')
    args = parser.parse_args()
    
    with psycopg2.connect(database=args.db, user=args.user) as connection:
        for scenario_id, scenario in enumerate(nminer.eval.bench.scenarios):
            print(f'Processing scenario {scenario_id}')
            
            cmp_col = scenario['cmp_col']
            with connection.cursor() as cursor:
                table = scenario['general']['table']
                sql = f'select distinct {cmp_col} from {table} ' \
                    f'where {cmp_col} is not null'
                cursor.execute(sql)
                cmp_vals = [v[0] for v in cursor.fetchall()]
                
                if scenario_id == 1:
                    languages = set()
                    for v in cmp_vals:
                        if v is not None:
                            for s in v.split(';'):
                                languages.add(s)
                    preds = [f"{cmp_col} like '%{nminer.sql.pred.sql_esc(l)}%'" 
                             for l in languages]
                else:
                    preds = [nminer.sql.pred.pred_sql(cmp_col, v) for v in cmp_vals]
                
                scenario['predicates'] = preds
                print(f'Extracted {len(preds)} predicates.')
                json_path = f'bench/batch{scenario_id}.json'
                with open(json_path, 'w') as file:
                    json.dump(scenario, file)


'''
Created on May 4, 2023

@author: immanueltrummer
'''
from setuptools import setup, find_packages

setup(
    name='naturalminer',
    version='0.1.0',
    description='Mine data for patterns described in natural language',
    url='https://github.com/itrummer/NaturalMiner',
    author='Immanuel Trummer',
    author_email='immanuel.trummer@gmail.com',
    packages=find_packages('src'),
    package_dir={'':'src'},
    install_requires=[
        'psycopg2-binary>=2.8, <3',
        'jinja2>=3.0, <4',
        'transformers>=4.9, <5',
        'sentence-transformers>=2.0, <3',
        'psycopg2-binary>=2.9.6, <3',
        'stable-baselines3[extra]>=2.0.0a4, <3'
        ],
    extras_require={
        'gui':['streamlit>=1.9, <=2.0']
        },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: GPU',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Topic :: Database :: Front-Ends',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
    ],
)


'''
Created on Apr 25, 2022

@author: immanueltrummer
'''
import json
import time
import datetime
from googleapiclient import errors

from google.cloud import storage
from googleapiclient import discovery

PROJECT_ID = 'baboons-348402'
_OPTIMIZER_API_DOCUMENT_BUCKET = 'caip-optimizer-public'
_OPTIMIZER_API_DOCUMENT_FILE = 'api/ml_public_google_rest_v1.json'

def read_api_document():
    client = storage.Client(PROJECT_ID)
    bucket = client.get_bucket(_OPTIMIZER_API_DOCUMENT_BUCKET)
    blob = bucket.get_blob(_OPTIMIZER_API_DOCUMENT_FILE)
    return blob.download_as_string()

ml = discovery.build_from_document(service=read_api_document())
print('Successfully built the client.')


# Update to your username
USER = 'itrummer' #@param {type: 'string'}

# These will be automatically filled in.
STUDY_ID = '{}_study_{}'.format(
    USER, datetime.datetime.now().strftime(
        '%Y%m%d_%H%M%S')) #@param {type: 'string'}
REGION = 'us-central1'

def study_parent():
    return 'projects/{}/locations/{}'.format(
        PROJECT_ID, REGION)

def study_name(study_id):
    return 'projects/{}/locations/{}/studies/{}'.format(
        PROJECT_ID, REGION, study_id)

def trial_parent(study_id):
    return study_name(study_id)

def trial_name(study_id, trial_id):
    return 'projects/{}/locations/{}/studies/{}/trials/{}'.format(
        PROJECT_ID, REGION, study_id, trial_id)

def operation_name(operation_id):
    return 'projects/{}/locations/{}/operations/{}'.format(
        PROJECT_ID, REGION, operation_id)


print('USER: {}'.format(USER))
print('PROJECT_ID: {}'.format(PROJECT_ID))
print('REGION: {}'.format(REGION))
print('STUDY_ID: {}'.format(STUDY_ID))


# Parameter Configuration
param_r = {
    'parameter': 'r',
    'type' : 'DOUBLE',
    'double_value_spec' : {
        'min_value' : 0,
        'max_value' : 1
    }
}

param_theta = {
    'parameter': 'theta',
    'type' : 'DOUBLE',
    'double_value_spec' : {
        'min_value' : 0,
        'max_value' : 1.57
    }
}

# Objective Metrics
metric_y1 = {
    'metric' : 'y1',
    'goal' : 'MINIMIZE'
}

metric_y2 = {
    'metric' : 'y2',
    'goal' : 'MAXIMIZE'
}

# Put it all together in a study configuration
study_config = {
    'algorithm' : 'ALGORITHM_UNSPECIFIED',  # Let the service choose the `default` algorithm.
    'parameters' : [param_r, param_theta,],
    'metrics' : [metric_y1, metric_y2,],
}

study = {'study_config': study_config}
print(json.dumps(study, indent=2, sort_keys=True))


# Creates a study
req = ml.projects().locations().studies().create(
    parent=study_parent(), studyId=STUDY_ID, body=study)
try :
    print(req.execute())
except errors.HttpError as e:
    if e.resp.status == 409:
        print('Study already existed.')
    else:
        raise e


import math


# r * sin(theta)
def Metric1Evaluation(r, theta):
    """Evaluate the first metric on the trial."""
    return r * math.sin(theta)


# r * cose(theta)
def Metric2Evaluation(r, theta):
    """Evaluate the second metric on the trial."""
    return r * math.cos(theta)


def CreateMeasurement(trial_id, r, theta):
    print(("=========== Start Trial: [{0}] =============").format(trial_id))
    
    # Evaluate both objective metrics for this trial
    y1 = Metric1Evaluation(r, theta)
    y2 = Metric2Evaluation(r, theta)
    print('[r = {0}, theta = {1}] => y1 = r*sin(theta) = {2}, y2 = r*cos(theta) = {3}'.format(r, theta, y1, y2))
    metric1 = {'metric': 'y1', 'value': y1}
    metric2 = {'metric': 'y2', 'value': y2}
    
    # Return the results for this trial
    measurement = {'step_count': 1, 'metrics': [metric1, metric2,]}
    return measurement


client_id = 'client1' #@param {type: 'string'}
suggestion_count_per_request =  5 #@param {type: 'integer'}
max_trial_id_to_stop =  50 #@param {type: 'integer'}

print('client_id: {}'.format(client_id))
print('suggestion_count_per_request: {}'.format(suggestion_count_per_request))
print('max_trial_id_to_stop: {}'.format(max_trial_id_to_stop))


trial_id = 0
while trial_id < max_trial_id_to_stop:
    # Requests trials.
    resp = ml.projects().locations().studies().trials().suggest(
      parent=trial_parent(STUDY_ID),
      body={'client_id': client_id, 'suggestion_count': suggestion_count_per_request}).execute()
    op_id = resp['name'].split('/')[-1]
    
    # Polls the suggestion long-running operations.
    get_op = ml.projects().locations().operations().get(name=operation_name(op_id))
    while True:
        operation = get_op.execute()
        if 'done' in operation and operation['done']:
            break
        time.sleep(1)
    
    for suggested_trial in get_op.execute()['response']['trials']:
        trial_id = int(suggested_trial['name'].split('/')[-1])
        # Featches the suggested trials.
        locations = ml.projects().locations()
        trial = locations.studies().trials().get(name=trial_name(STUDY_ID, trial_id)).execute()
        if trial['state'] in ['COMPLETED', 'INFEASIBLE']:
            continue
        
    params = {}
    for param in trial['parameters']:
        if param['parameter'] == 'r':
            r = param['floatValue']
        elif param['parameter'] == 'theta':
            theta = param['floatValue']
    
    # Evaluates trials and reports measurement.
    ml.projects().locations().studies().trials().addMeasurement(
        name=trial_name(STUDY_ID, trial_id),
        body={'measurement': CreateMeasurement(trial_id, r, theta)}).execute()
    # Completes the trial.
    ml.projects().locations().studies().trials().complete(
        name=trial_name(STUDY_ID, trial_id)).execute()



max_trials_to_annotate = 20

import matplotlib.pyplot as plt
trial_ids = []
y1 = []
y2 = []
resp = ml.projects().locations().studies().trials().list(parent=trial_parent(STUDY_ID)).execute()
for trial in resp['trials']:
    if 'finalMeasurement' in trial:
        trial_ids.append(int(trial['name'].split('/')[-1]))
        metrics = trial['finalMeasurement']['metrics']
        try:
            y1.append([m for m in metrics if m['metric'] == "y1"][0]['value'])
            y2.append([m for m in metrics if m['metric'] == "y2"][0]['value'])
        except:
            pass

fig, ax = plt.subplots()
ax.scatter(y1, y2)
plt.xlabel("y1=r*sin(theta)")
plt.ylabel("y2=r*cos(theta)");
for i, trial_id in enumerate(trial_ids):
    # Only annotates the last `max_trials_to_annotate` trials
    if i > len(trial_ids) - max_trials_to_annotate:
        try:
            ax.annotate(trial_id, (y1[i], y2[i]))
        except:
            pass
plt.gcf().set_size_inches((16, 16))